import datetime as dt
import os
import tempfile
import zipfile
from dataclasses import astuple, replace
from pathlib import Path

import numpy as np
import pytest
import xarray as xr

from ravenpy.config.commands import (
    ChannelProfileCommand,
    GriddedForcingCommand,
    GridWeightsCommand,
    HRUStateVariableTableCommand,
    LandUseClassesCommand,
    ObservationDataCommand,
    SBGroupPropertyMultiplierCommand,
    SoilClassesCommand,
    SoilProfilesCommand,
    Sub,
    VegetationClassesCommand,
)
from ravenpy.config.rvs import RVI
from ravenpy.models import (
    GR4JCN,
    GR4JCN_OST,
    HBVEC,
    HBVEC_OST,
    HMETS,
    HMETS_OST,
    MOHYSE,
    MOHYSE_OST,
    Raven,
    RavenError,
    get_average_annual_runoff,
)
from ravenpy.utilities.testdata import get_local_testdata

from .common import _convert_2d

TS = get_local_testdata(
    "raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc"
)

# Link to THREDDS Data Server netCDF testdata
TDS = "https://pavics.ouranos.ca/twitcher/ows/proxy/thredds/dodsC/birdhouse/testdata/raven"


@pytest.fixture
def input2d(tmpdir):
    """Convert 1D input to 2D output by copying all the time series along a new region dimension."""
    ds = _convert_2d(TS)
    fn_out = os.path.join(tmpdir, "input2d.nc")
    ds.to_netcdf(fn_out)
    return Path(fn_out)


def test_race():
    model1 = GR4JCN()
    model1.config.rvi.suppress_output = True
    model2 = GR4JCN()
    ost = GR4JCN_OST()

    assert model1.config.rvi.suppress_output.startswith(":SuppressOutput")
    assert model2.config.rvi.suppress_output == ""
    assert ost.config.rvi.suppress_output.startswith(":SuppressOutput")


# Salmon catchment is now split into land- and lake-part.
# The areas do not sum up to overall area of 4250.6 [km2].
# This is the reason the "test_routing" will give differnt
# results compared to "test_simple". The "salmon_land_hru"
# however is kept at the overall area of 4250.6 [km2] such
# that other tests still obtain same results as before.
salmon_land_hru_1 = dict(
    area=4250.6, elevation=843.0, latitude=54.4848, longitude=-123.3659
)
salmon_lake_hru_1 = dict(area=100.0, elevation=839.0, latitude=54.0, longitude=-123.4)
salmon_land_hru_2 = dict(
    area=2000.0, elevation=835.0, latitude=54.123, longitude=-123.4234
)


class TestGR4JCN:
    def test_error(self):
        model = GR4JCN()

        model.config.rvp.params = model.Params(
            0.529, -3.396, 407.29, 1.072, 16.9, 0.947
        )

        with pytest.raises(RavenError) as exc:
            model(TS)

        assert "CHydroUnit constructor:: HRU 1 has a negative or zero area" in str(
            exc.value
        )

    def test_simple(self):
        model = GR4JCN()

        model.config.rvi.start_date = dt.datetime(2000, 1, 1)
        model.config.rvi.end_date = dt.datetime(2002, 1, 1)
        model.config.rvi.run_name = "test"

        model.config.rvh.hrus = (GR4JCN.LandHRU(**salmon_land_hru_1),)

        model.config.rvp.params = model.Params(
            0.529, -3.396, 407.29, 1.072, 16.9, 0.947
        )

        total_area_in_m2 = model.config.rvh.hrus[0].area * 1000 * 1000
        model.config.rvp.avg_annual_runoff = get_average_annual_runoff(
            TS, total_area_in_m2
        )

        np.testing.assert_almost_equal(
            model.config.rvp.avg_annual_runoff, 208.4805694844741
        )

        assert model.config.rvi.suppress_output == ""

        model(TS)

        # ------------
        # Check quality (diagnostic) of simulated streamflow values
        # ------------
        d = model.diagnostics
        np.testing.assert_almost_equal(d["DIAG_NASH_SUTCLIFFE"], -0.117301, 4)

        # ------------
        # Check simulated streamflow values q_sim
        # ------------
        hds = model.q_sim

        assert hds.attrs["long_name"] == "Simulated outflows"

        assert len(hds.nbasins) == 1  # number of "gauged" basins is 1

        # We only have one SB with gauged=True, so the output has a single column.
        # The number of time steps simulated between (2000, 1, 1) and
        # (2002, 1, 1) is 732.
        assert hds.shape == (732, 1)

        # Check simulated streamflow at first three timesteps and three simulated
        # timesteps in the middle of the simulation period.
        dates = (
            "2000-01-01",
            "2000-01-02",
            "2000-01-03",
            "2001-01-30",
            "2001-01-31",
            "2001-02-01",
        )

        target_q_sim = [0.0, 0.165788, 0.559366, 12.374606, 12.33398, 12.293458]

        for t in range(6):
            np.testing.assert_almost_equal(
                hds.sel(nbasins=0, time=dates[t]), target_q_sim[t], 4
            )

        # ------------
        # Check parser
        # ------------

        assert model.config.rvi.calendar == RVI.CalendarOptions.GREGORIAN.value

        # ------------
        # Check saved HRU states saved in RVC
        # ------------
        assert 1 in model.solution.hru_states

        # ------------
        # Check attributes
        # ------------
        assert model.hydrograph.attrs["model_id"] == "gr4jcn"

    def test_routing(self):
        """We need at least 2 subbasins to activate routing."""
        model = GR4JCN()

        ts_2d = get_local_testdata(
            "raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily_2d.nc"
        )

        #########
        # R V I #
        #########

        model.config.rvi.start_date = dt.datetime(2000, 1, 1)
        model.config.rvi.end_date = dt.datetime(2002, 1, 1)
        model.config.rvi.run_name = "test_gr4jcn_routing"
        model.config.rvi.routing = "ROUTE_DIFFUSIVE_WAVE"

        #########
        # R V H #
        #########

        # Here we assume that we have two subbasins. The first one (subbasin_id=10)
        # has a lake (hru_id=2; area-100km2) and the rest is covered by land (hru_id=1;
        # area=4250.6km2). The second subbasin (subbasin_id=20) does not contain a
        # lake and is hence only land (hru_id=3; area=2000km2).
        #
        # Later the routing product will tell us which basin flows into which. Here
        # we assume that the first subbasin (subbasin_id=10) drains into the second
        # (subbasin_id=20). At the outlet of this second one we have an observation
        # station (see :ObservationData in RVT). We will compare these observations
        # with the simulated streamflow. That is the reason why "gauged=True" for
        # the second basin.

        # HRU IDs are 1 to 3
        model.config.rvh.hrus = (
            GR4JCN.LandHRU(hru_id=1, subbasin_id=10, **salmon_land_hru_1),
            GR4JCN.LakeHRU(hru_id=2, subbasin_id=10, **salmon_lake_hru_1),
            GR4JCN.LandHRU(hru_id=3, subbasin_id=20, **salmon_land_hru_2),
        )

        # Sub-basin IDs are 10 and 20 (not 1 and 2), to help disambiguate
        model.config.rvh.subbasins = (
            # gauged = False:
            # Usually this output would only be written for user's convenience.
            # There is usually no observation of streamflow available within
            # catchments; only at the outlet. That's most commonly the reason
            # why a catchment is defined as it is defined.
            Sub(
                name="upstream",
                subbasin_id=10,
                downstream_id=20,
                profile="chn_10",
                gauged=False,
            ),
            # gauged = True:
            # Since this is the outlet, this would usually be what we calibrate
            # against (i.e. we try to match this to Qobs).
            Sub(
                name="downstream",
                subbasin_id=20,
                downstream_id=-1,
                profile="chn_20",
                gauged=True,
            ),
        )

        model.config.rvh.land_subbasin_property_multiplier = (
            SBGroupPropertyMultiplierCommand("Land", "MANNINGS_N", 1.0)
        )
        model.config.rvh.lake_subbasin_property_multiplier = (
            SBGroupPropertyMultiplierCommand("Lakes", "RESERVOIR_CREST_WIDTH", 1.0)
        )

        #########
        # R V T #
        #########

        gws = GridWeightsCommand(
            number_hrus=3,
            number_grid_cells=1,
            # Here we have a special case: station is 0 for every row because the example NC
            # has only one region/station (which is column 0)
            data=((1, 0, 1.0), (2, 0, 1.0), (3, 0, 1.0)),
        )
        # These will be shared (inline) to all the StationForcing commands in the RVT
        model.config.rvt.grid_weights = gws

        #########
        # R V P #
        #########

        model.config.rvp.params = model.Params(
            0.529, -3.396, 407.29, 1.072, 16.9, 0.947
        )

        total_area_in_km2 = sum(hru.area for hru in model.config.rvh.hrus)
        total_area_in_m2 = total_area_in_km2 * 1000 * 1000
        model.config.rvp.avg_annual_runoff = get_average_annual_runoff(
            ts_2d, total_area_in_m2
        )

        np.testing.assert_almost_equal(
            model.config.rvp.avg_annual_runoff, 139.5407534171111
        )

        # These channel profiles describe the geometry of the actual river crossection.
        # The eight points (x) to describe the following geometry are given in each
        # profile:
        #
        # ----x                                     x---
        #      \           FLOODPLAIN             /
        #       x----x                     x----x
        #             \                  /
        #               \   RIVERBED   /
        #                 x----------x
        #
        model.config.rvp.channel_profiles = [
            ChannelProfileCommand(
                name="chn_10",
                bed_slope=7.62066e-05,
                survey_points=[
                    (0, 463.647),
                    (16.0, 459.647),
                    (90.9828, 459.647),
                    (92.9828, 458.647),
                    (126.4742, 458.647),
                    (128.4742, 459.647),
                    (203.457, 459.647),
                    (219.457, 463.647),
                ],
                roughness_zones=[
                    (0, 0.0909167),
                    (90.9828, 0.035),
                    (128.4742, 0.0909167),
                ],
            ),
            ChannelProfileCommand(
                name="chn_20",
                bed_slope=9.95895e-05,
                survey_points=[
                    (0, 450.657),
                    (16.0, 446.657),
                    (85.0166, 446.657),
                    (87.0166, 445.657),
                    (117.5249, 445.657),
                    (119.5249, 446.657),
                    (188.54149999999998, 446.657),
                    (204.54149999999998, 450.657),
                ],
                roughness_zones=[
                    (0, 0.0915769),
                    (85.0166, 0.035),
                    (119.5249, 0.0915769),
                ],
            ),
        ]

        #############
        # Run model #
        #############

        model(ts_2d)

        ###########
        # Verify  #
        ###########

        hds = model.q_sim

        assert len(hds.nbasins) == 1  # number of "gauged" basins is 1

        # We only have one SB with gauged=True, so the output has a single column.
        # The number of time steps simulated between (2000, 1, 1) and
        # (2002, 1, 1) is 732.
        assert hds.shape == (732, 1)

        # Check simulated streamflow at first three timesteps and three simulated
        # timesteps in the middle of the simulation period.
        dates = (
            "2000-01-01",
            "2000-01-02",
            "2000-01-03",
            "2001-01-30",
            "2001-01-31",
            "2001-02-01",
        )

        target_q_sim = [0.0, 0.304073, 0.980807, 17.54049, 17.409493, 17.437954]

        for t in range(6):
            np.testing.assert_almost_equal(
                hds.sel(nbasins=0, time=dates[t]), target_q_sim[t], 4
            )

        # For lumped GR4J model we have 1 subbasin and 1 HRU as well as no routing, no
        # channel profiles, and the area of the entire basin is 4250.6 [km2]. Comparison
        # of simulated and observed streamflow at outlet yielded:
        # np.testing.assert_almost_equal(d["DIAG_NASH_SUTCLIFFE"], -0.116971, 4)
        #
        # This is now a different value due to:
        # - basin we have here is larger (4250.6 [km2] + 100 [km2] + 2000.0 [km2])
        # - we do routing: so water from subbasin 1 needs some time to arrive at the
        #   outlet of subbasin 2
        d = model.diagnostics
        np.testing.assert_almost_equal(d["DIAG_NASH_SUTCLIFFE"], -0.0141168, 4)

    def test_config_update(self):
        model = GR4JCN()

        # This is a regular attribute member
        model.config.update("run_name", "test")
        assert model.config.rvi.run_name == "test"

        # This is a computed property
        model.config.update("evaporation", "PET_FROMMONTHLY")
        assert model.config.rvi.evaporation == "PET_FROMMONTHLY"

        # Existing property but wrong value (the enum cast should throw an error)
        with pytest.raises(ValueError):
            model.config.update("routing", "WRONG")

        # Non-existing attribute
        with pytest.raises(AttributeError):
            model.config.update("why", "not?")

        # Params

        model.config.update(
            "params", np.array([0.529, -3.396, 407.29, 1.072, 16.9, 0.947])
        )
        assert model.config.rvp.params.GR4J_X1 == 0.529

        model.config.update("params", [0.529, -3.396, 407.29, 1.072, 16.9, 0.947])
        assert model.config.rvp.params.GR4J_X1 == 0.529

        model.config.update("params", (0.529, -3.396, 407.29, 1.072, 16.9, 0.947))
        assert model.config.rvp.params.GR4J_X1 == 0.529

    def test_run(self):
        model = GR4JCN()

        model.config.rvh.hrus = (GR4JCN.LandHRU(**salmon_land_hru_1),)

        model(
            TS,
            area=4250.6,
            elevation=843.0,
            latitude=54.4848,
            longitude=-123.3659,
            start_date=dt.datetime(2000, 1, 1),
            end_date=dt.datetime(2002, 1, 1),
            params=(0.529, -3.396, 407.29, 1.072, 16.9, 0.947),
            suppress_output=False,
        )
        d = model.diagnostics

        np.testing.assert_almost_equal(d["DIAG_NASH_SUTCLIFFE"], -0.117301, 4)

    def test_run_new_hrus_param(self):
        model = GR4JCN()

        model(
            TS,
            start_date=dt.datetime(2000, 1, 1),
            end_date=dt.datetime(2002, 1, 1),
            params=(0.529, -3.396, 407.29, 1.072, 16.9, 0.947),
            suppress_output=False,
            hrus=(GR4JCN.LandHRU(**salmon_land_hru_1),),
        )
        d = model.diagnostics

        np.testing.assert_almost_equal(d["DIAG_NASH_SUTCLIFFE"], -0.117301, 4)

    # @pytest.mark.skip
    def test_overwrite(self):
        model = GR4JCN()

        model.config.rvh.hrus = (GR4JCN.LandHRU(**salmon_land_hru_1),)

        model(
            TS,
            start_date=dt.datetime(2000, 1, 1),
            end_date=dt.datetime(2002, 1, 1),
            params=(0.529, -3.396, 407.29, 1.072, 16.9, 0.947),
        )
        assert model.config.rvi.suppress_output == ""

        qsim1 = model.q_sim.copy(deep=True)
        m1 = qsim1.mean()

        # This is only needed temporarily while we fix this: https://github.com/CSHS-CWRA/RavenPy/issues/4
        # Please remove when fixed!
        model.hydrograph.close()  # Needed with xarray 0.16.1

        model(TS, params=(0.5289, -3.397, 407.3, 1.071, 16.89, 0.948), overwrite=True)

        qsim2 = model.q_sim.copy(deep=True)
        m2 = qsim2.mean()

        # This is only needed temporarily while we fix this: https://github.com/CSHS-CWRA/RavenPy/issues/4
        # Please remove when fixed!
        model.hydrograph.close()  # Needed with xarray 0.16.1

        assert m1 != m2

        np.testing.assert_almost_equal(m1, m2, 1)

        d = model.diagnostics

        np.testing.assert_almost_equal(d["DIAG_NASH_SUTCLIFFE"], -0.117315, 4)

        model.config.rvc.hru_states[1] = HRUStateVariableTableCommand.Record(soil0=0)

        # Set initial conditions explicitly
        model(
            TS,
            end_date=dt.datetime(2001, 2, 1),
            # hru_state=HRUStateVariableTableCommand.Record(soil0=0),
            overwrite=True,
        )
        assert model.q_sim.isel(time=1).values[0] < qsim2.isel(time=1).values[0]

    def test_resume(self):
        model_ab = GR4JCN()
        model_ab.config.rvh.hrus = (GR4JCN.LandHRU(**salmon_land_hru_1),)
        kwargs = dict(
            params=(0.529, -3.396, 407.29, 1.072, 16.9, 0.947),
        )
        # Reference run
        model_ab(
            TS,
            run_name="run_ab",
            start_date=dt.datetime(2000, 1, 1),
            end_date=dt.datetime(2001, 1, 1),
            **kwargs,
        )

        model_a = GR4JCN()

        model_a.config.rvh.hrus = (GR4JCN.LandHRU(**salmon_land_hru_1),)
        model_a(
            TS,
            run_name="run_a",
            start_date=dt.datetime(2000, 1, 1),
            end_date=dt.datetime(2000, 7, 1),
            **kwargs,
        )

        # Path to solution file from run A
        rvc = model_a.outputs["solution"]

        # Resume with final state from live model
        model_a.resume()

        model_a(
            TS,
            run_name="run_2",
            start_date=dt.datetime(2000, 7, 1),
            end_date=dt.datetime(2001, 1, 1),
            **kwargs,
        )

        for key in ["Soil Water[0]", "Soil Water[1]"]:
            np.testing.assert_array_almost_equal(
                model_a.storage[key] - model_ab.storage[key], 0, 5
            )

        # Resume with final state from saved solution file
        model_b = GR4JCN()
        model_b.config.rvh.hrus = (GR4JCN.LandHRU(**salmon_land_hru_1),)
        model_b.resume(
            rvc
        )  # <--------- And this is how you feed it to a brand new model.
        model_b(
            TS,
            run_name="run_2",
            start_date=dt.datetime(2000, 7, 1),
            end_date=dt.datetime(2001, 1, 1),
            **kwargs,
        )

        for key in ["Soil Water[0]", "Soil Water[1]"]:
            np.testing.assert_array_almost_equal(
                model_b.storage[key] - model_ab.storage[key], 0, 5
            )

        # model.solution loads the solution in a dictionary. I expected the variables to be identical,
        # but some atmosphere related attributes are way off. Is it possible that `ATMOSPHERE` and `ATMOS_PRECIP` are
        # cumulative sums of precipitation over the run ?
        # assert model_b.solution == model_ab.solution # This does not work. Atmosphere attributes are off.

    def test_resume_earlier(self):
        """Check that we can resume a run with the start date set at another date than the time stamp in the
        solution."""
        params = (0.529, -3.396, 407.29, 1.072, 16.9, 0.947)
        # Reference run
        model = GR4JCN()
        model.config.rvh.hrus = (GR4JCN.LandHRU(**salmon_land_hru_1),)
        model(
            TS,
            run_name="run_a",
            start_date=dt.datetime(2000, 1, 1),
            end_date=dt.datetime(2000, 2, 1),
            params=params,
        )

        s_a = model.storage["Soil Water[0]"].isel(time=-1)

        # Path to solution file from run A
        rvc = model.outputs["solution"]

        # Resume with final state from live model
        # We have two options to do this:
        # 1. Replace model template by solution file as is: model.resume()
        # 2. Replace variable in RVC class by parsed values: model.rvc.parse(rvc.read_text())
        # I think in many cases option 2 will prove simpler.

        model.config.rvc.parse_solution(rvc.read_text())

        model(
            TS,
            run_name="run_b",
            start_date=dt.datetime(2000, 1, 1),
            end_date=dt.datetime(2000, 2, 1),
            params=params,
        )

        s_b = model.storage["Soil Water[0]"].isel(time=-1)
        assert s_a != s_b

    def test_update_soil_water(self):
        params = (0.529, -3.396, 407.29, 1.072, 16.9, 0.947)
        # Reference run
        model = GR4JCN()
        model.config.rvh.hrus = (GR4JCN.LandHRU(**salmon_land_hru_1),)
        model(
            TS,
            run_name="run_a",
            start_date=dt.datetime(2000, 1, 1),
            end_date=dt.datetime(2000, 2, 1),
            params=params,
        )

        s_0 = float(model.storage["Soil Water[0]"].isel(time=-1).values)
        s_1 = float(model.storage["Soil Water[1]"].isel(time=-1).values)

        # hru_state = replace(model.rvc.hru_state, soil0=s_0, soil1=s_1)
        model.config.rvc.hru_states[1] = replace(
            model.config.rvc.hru_states[1], soil0=s_0, soil1=s_1
        )

        model(
            TS,
            run_name="run_b",
            start_date=dt.datetime(2000, 1, 1),
            end_date=dt.datetime(2000, 2, 1),
            # hru_state=hru_state,
            params=params,
        )

        assert s_0 != model.storage["Soil Water[0]"].isel(time=-1)
        assert s_1 != model.storage["Soil Water[1]"].isel(time=-1)

    def test_version(self):
        model = Raven()
        assert model.version == "3.0.4"

        model = GR4JCN()
        assert model.version == "3.0.4"

    def test_parallel_params(self):
        model = GR4JCN()
        model.config.rvh.hrus = (GR4JCN.LandHRU(**salmon_land_hru_1),)

        model(
            TS,
            start_date=dt.datetime(2000, 1, 1),
            end_date=dt.datetime(2002, 1, 1),
            params=[
                (0.529, -3.396, 407.29, 1.072, 16.9, 0.947),
                (0.528, -3.4, 407.3, 1.07, 17, 0.95),
            ],
            suppress_output=False,
        )

        assert len(model.diagnostics) == 2
        assert model.hydrograph.dims["params"] == 2
        z = zipfile.ZipFile(model.outputs["rv_config"])
        assert len(z.filelist) == 10

    def test_parallel_basins(self, input2d):
        ts = input2d
        model = GR4JCN()
        model.config.rvh.hrus = (GR4JCN.LandHRU(**salmon_land_hru_1),)

        model(
            ts,
            start_date=dt.datetime(2000, 1, 1),
            end_date=dt.datetime(2002, 1, 1),
            params=[0.529, -3.396, 407.29, 1.072, 16.9, 0.947],
            nc_index=[0, 0],
            # name=["basin1", "basin2"],  # Not sure about this..
            suppress_output=False,
        )

        assert len(model.diagnostics) == 2
        assert len(model.hydrograph.nbasins) == 2
        np.testing.assert_array_equal(
            model.hydrograph.basin_name[:], ["sub_001", "sub_001"]
        )
        z = zipfile.ZipFile(model.outputs["rv_config"])
        assert len(z.filelist) == 10

    @pytest.mark.online
    def test_dap(self):
        """Test Raven with DAP link instead of local netCDF file."""
        model = GR4JCN()
        config = dict(
            start_date=dt.datetime(2000, 6, 1),
            end_date=dt.datetime(2000, 6, 10),
            run_name="test",
            hrus=(GR4JCN.LandHRU(**salmon_land_hru_1),),
            params=model.Params(0.529, -3.396, 407.29, 1.072, 16.9, 0.947),
        )

        ts = (
            f"{TDS}/raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc"
        )
        model(ts, **config)

    @pytest.mark.online
    def test_canopex(self):
        CANOPEX_DAP = (
            "https://pavics.ouranos.ca/twitcher/ows/proxy/thredds/dodsC/birdhouse/ets"
            "/Watersheds_5797_cfcompliant.nc"
        )
        model = GR4JCN()
        config = dict(
            start_date=dt.datetime(2010, 6, 1),
            end_date=dt.datetime(2010, 6, 10),
            nc_index=5600,
            run_name="Test_run",
            rain_snow_fraction="RAINSNOW_DINGMAN",
            tasmax={"offset": -273.15},
            tasmin={"offset": -273.15},
            pr={"scale": 86400.0},
            hrus=[
                model.LandHRU(
                    area=3650.47, latitude=49.51, longitude=-95.72, elevation=330.59
                )
            ],
            params=model.Params(108.02, 2.8693, 25.352, 1.3696, 1.2483, 0.30679),
        )
        model(ts=CANOPEX_DAP, **config)


class TestGR4JCN_OST:
    def test_simple(self):
        model = GR4JCN_OST()
        model.config.rvh.hrus = (GR4JCN.LandHRU(**salmon_land_hru_1),)
        params = (0.529, -3.396, 407.29, 1.072, 16.9, 0.053)
        low = (0.01, -15.0, 10.0, 0.0, 1.0, 0.0)
        high = (2.5, 10.0, 700.0, 7.0, 30.0, 1.0)

        model.configure(
            get_local_testdata("ostrich-gr4j-cemaneige/OstRandomNumbers.txt")
        )

        model(
            TS,
            start_date=dt.datetime(1954, 1, 1),
            duration=208,
            params=params,
            lowerBounds=low,
            upperBounds=high,
            algorithm="DDS",
            random_seed=0,
            max_iterations=10,
        )

        d = model.diagnostics

        np.testing.assert_almost_equal(d["DIAG_NASH_SUTCLIFFE"], 0.50717, 4)

        # Random number seed: 123
        # Budget:             10
        # Algorithm:          DDS
        # :StartDate          1954-01-01 00:00:00
        # :Duration           208
        opt_para = astuple(model.calibrated_params)
        opt_func = model.obj_func

        np.testing.assert_almost_equal(
            opt_para,
            [2.424726, 3.758972, 204.3856, 5.866946, 16.60408, 0.3728098],
            4,
            err_msg="calibrated parameter set is not matching expected value",
        )

        np.testing.assert_almost_equal(
            opt_func,
            -0.50717,
            4,
            err_msg="calibrated NSE is not matching expected value",
        )

        # # Random number seed: 123
        # # Budget:             50
        # # Algorithm:          DDS
        # # :StartDate          1954-01-01 00:00:00
        # # :Duration           20819
        # np.testing.assert_almost_equal( opt_para, [0.3243268,3.034247,407.2890,2.722774,12.18124,0.9468769], 4,
        #                                 err_msg='calibrated parameter set is not matching expected value')
        # np.testing.assert_almost_equal( opt_func, -0.5779910, 4,
        #                                 err_msg='calibrated NSE is not matching expected value')
        gr4j = GR4JCN()
        gr4j.config.rvh.hrus = (GR4JCN.LandHRU(**salmon_land_hru_1),)

        gr4j(
            TS,
            start_date=dt.datetime(1954, 1, 1),
            duration=208,
            params=model.calibrated_params,
        )

        np.testing.assert_almost_equal(
            gr4j.diagnostics["DIAG_NASH_SUTCLIFFE"], d["DIAG_NASH_SUTCLIFFE"]
        )


class TestHMETS:
    def test_simple(self):
        model = HMETS()
        params = (
            9.5019,
            0.2774,
            6.3942,
            0.6884,
            1.2875,
            5.4134,
            2.3641,
            0.0973,
            0.0464,
            0.1998,
            0.0222,
            -1.0919,
            2.6851,
            0.3740,
            1.0000,
            0.4739,
            0.0114,
            0.0243,
            0.0069,
            310.7211,
            916.1947,
        )

        model(
            TS,
            start_date=dt.datetime(2000, 1, 1),
            end_date=dt.datetime(2002, 1, 1),
            area=4250.6,
            elevation=843.0,
            latitude=54.4848,
            longitude=-123.3659,
            params=params,
            suppress_output=True,
        )

        d = model.diagnostics

        np.testing.assert_almost_equal(d["DIAG_NASH_SUTCLIFFE"], -3.0132, 4)


class TestHMETS_OST:
    def test_simple(self):
        model = HMETS_OST()

        model.configure(
            get_local_testdata("ostrich-gr4j-cemaneige/OstRandomNumbers.txt")
        )

        params = (
            9.5019,
            0.2774,
            6.3942,
            0.6884,
            1.2875,
            5.4134,
            2.3641,
            0.0973,
            0.0464,
            0.1998,
            0.0222,
            -1.0919,
            2.6851,
            0.3740,
            1.0000,
            0.4739,
            0.0114,
            0.0243,
            0.0069,
            310.7211,
            916.1947,
        )
        low = (
            0.3,
            0.01,
            0.5,
            0.15,
            0.0,
            0.0,
            -2.0,
            0.01,
            0.0,
            0.01,
            0.005,
            -5.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.00001,
            0.0,
            0.00001,
            0.0,
            0.0,
        )
        high = (
            20.0,
            5.0,
            13.0,
            1.5,
            20.0,
            20.0,
            3.0,
            0.2,
            0.1,
            0.3,
            0.1,
            2.0,
            5.0,
            1.0,
            3.0,
            1.0,
            0.02,
            0.1,
            0.01,
            0.5,
            2.0,
        )

        model(
            TS,
            start_date=dt.datetime(1954, 1, 1),
            duration=208,
            area=4250.6,
            elevation=843.0,
            latitude=54.4848,
            longitude=-123.3659,
            params=params,
            lowerBounds=low,
            upperBounds=high,
            algorithm="DDS",
            random_seed=0,
            max_iterations=10,
        )

        d = model.diagnostics

        np.testing.assert_almost_equal(d["DIAG_NASH_SUTCLIFFE"], -1.43474, 4)

        opt_para = model.optimized_parameters
        opt_func = model.obj_func

        # # Random number seed: 123
        # # Budget:             50
        # # Algorithm:          DDS
        # # :StartDate          1954-01-01 00:00:00
        # # :Duration           20819
        # np.testing.assert_almost_equal( opt_para, [0.3243268,3.034247,407.2890,2.722774,12.18124,0.9468769], 4,
        #                                 err_msg='calibrated parameter set is not matching expected value')
        # np.testing.assert_almost_equal( opt_func, -0.5779910, 4,
        #                                 err_msg='calibrated NSE is not matching expected value')
        #
        # Random number seed: 123                         #
        # Budget:             10                          #      This is the setup used for testing:
        # Algorithm:          DDS                         #         shorter sim-period and lower budget
        # :StartDate          1954-01-01 00:00:00         #      First tested that example below matches
        # :Duration           208                         #

        expected_value = [
            1.777842e01,
            3.317211e00,
            5.727342e00,
            1.419491e00,
            1.382141e01,
            1.637954e01,
            7.166296e-01,
            1.389346e-01,
            2.620464e-02,
            2.245525e-01,
            2.839426e-02,
            -2.003810e00,
            9.479623e-01,
            4.803857e-01,
            2.524914e00,
            4.117232e-01,
            1.950058e-02,
            4.494123e-02,
            1.405815e-03,
            2.815803e-02,
            1.007823e00,
        ]
        np.testing.assert_almost_equal(
            opt_para,
            expected_value,
            4,
            err_msg="calibrated parameter set is not matching expected value",
        )
        np.testing.assert_almost_equal(
            opt_func,
            1.43474,
            4,
            err_msg="calibrated NSE is not matching expected value",
        )

        # # Random number seed: 123                       #
        # # Budget:             50                        #      This is the setup in the Wiki:
        # # Algorithm:          DDS                       #      https://github.com/Ouranosinc/raven/wiki/
        # # :StartDate          1954-01-01 00:00:00       #      Technical-Notes#example-setups-for-hmets
        # # :Duration           20819                     #
        # np.testing.assert_almost_equal(opt_para, [5.008045E+00, 7.960246E-02, 4.332698E+00, 4.978125E-01,
        #                                           1.997029E+00, 6.269773E-01, 1.516961E+00, 8.180383E-02,
        #                                           6.730663E-02, 2.137822E-02, 2.097163E-02, 1.773348E+00,
        #                                           3.036039E-01, 1.928524E-02, 1.758471E+00, 8.942299E-01,
        #                                           8.741980E-03, 5.036474E-02, 9.465804E-03, 1.851839E-01,
        #                                           1.653934E-01, 2.624006E+00, 8.868485E-02, 9.259195E+01,
        #                                           8.269670E+01], 4,
        #                                err_msg='calibrated parameter set is not matching expected value')
        # np.testing.assert_almost_equal(opt_func, -6.350490E-01, 4,
        #                                err_msg='calibrated NSE is not matching expected value')
        hmets = HMETS()
        hmets(
            TS,
            start_date=dt.datetime(1954, 1, 1),
            duration=208,
            area=4250.6,
            elevation=843.0,
            latitude=54.4848,
            longitude=-123.3659,
            params=model.calibrated_params,
        )

        np.testing.assert_almost_equal(
            hmets.diagnostics["DIAG_NASH_SUTCLIFFE"], d["DIAG_NASH_SUTCLIFFE"], 4
        )


class TestMOHYSE:
    def test_simple(self):
        model = MOHYSE()
        params = (
            1.0,
            0.0468,
            4.2952,
            2.658,
            0.4038,
            0.0621,
            0.0273,
            0.0453,
            0.9039,
            5.6167,
        )

        model(
            TS,
            start_date=dt.datetime(2000, 1, 1),
            end_date=dt.datetime(2002, 1, 1),
            area=4250.6,
            elevation=843.0,
            latitude=54.4848,
            longitude=-123.3659,
            params=params,
            suppress_output=True,
        )

        d = model.diagnostics
        np.testing.assert_almost_equal(d["DIAG_NASH_SUTCLIFFE"], 0.194612, 4)


class TestMOHYSE_OST:
    def test_simple(self):
        model = MOHYSE_OST()

        model.configure(
            get_local_testdata("ostrich-gr4j-cemaneige/OstRandomNumbers.txt")
        )

        params = (
            1.0,
            0.0468,
            4.2952,
            2.658,
            0.4038,
            0.0621,
            0.0273,
            0.0453,
            0.9039,
            5.6167,
        )

        low_p = (0.01, 0.01, 0.01, -5.00, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01)
        high_p = (20.0, 1.0, 20.0, 5.0, 0.5, 1.0, 1.0, 1.0, 15.0, 15.0)

        model(
            TS,
            start_date=dt.datetime(1954, 1, 1),
            duration=208,
            area=4250.6,
            elevation=843.0,
            latitude=54.4848,
            longitude=-123.3659,
            params=params,
            lowerBounds=low_p,
            upperBounds=high_p,
            algorithm="DDS",
            random_seed=0,
            max_iterations=10,
        )

        d = model.diagnostics
        np.testing.assert_almost_equal(d["DIAG_NASH_SUTCLIFFE"], 0.3826810, 4)

        opt_para = model.optimized_parameters
        opt_func = model.obj_func

        # # Random number seed: 123
        # # Budget:             50
        # # Algorithm:          DDS
        # # :StartDate          1954-01-01 00:00:00
        # # :Duration           20819
        # np.testing.assert_almost_equal( opt_para, [0.3243268,3.034247,407.2890,2.722774,12.18124,0.9468769], 4,
        #                                 err_msg='calibrated parameter set is not matching expected value')
        # np.testing.assert_almost_equal( opt_func, -0.5779910, 4,
        #                                 err_msg='calibrated NSE is not matching expected value')
        #
        # Random number seed: 123                         #
        # Budget:             10                          #      This is the setup used for testing:
        # Algorithm:          DDS                         #         shorter sim-period and lower budget
        # :StartDate          1954-01-01 00:00:00         #      First tested that example below matches
        # :Duration           208                         #
        np.testing.assert_almost_equal(
            opt_para,
            [
                7.721801e00,
                8.551484e-01,
                1.774571e01,
                1.627677e00,
                7.702450e-02,
                9.409600e-01,
                6.941596e-01,
                8.207870e-01,
                8.154455e00,
                1.018226e01,
            ],
            4,
            err_msg="calibrated parameter set is not matching expected value",
        )
        np.testing.assert_almost_equal(
            opt_func,
            -0.3826810,
            4,
            err_msg="calibrated NSE is not matching expected value",
        )

        # # Random number seed: 123                       #
        # # Budget:             50                        #      This is the setup in the Wiki:
        # # Algorithm:          DDS                       #      https://github.com/Ouranosinc/raven/wiki/
        # # :StartDate          1954-01-01 00:00:00       #      Technical-Notes#example-setups-for-mohyse
        # # :Duration           20819                     #
        # np.testing.assert_almost_equal(opt_para, [1.517286E+01, 7.112556E-01, 1.981243E+01, -4.193046E+00,
        #                                           1.791486E-01, 9.774897E-01, 5.353541E-01, 6.686806E-01,
        #                                           1.040908E+01, 1.132304E+01, 8.831552E-02], 4,
        #                                err_msg='calibrated parameter set is not matching expected value')
        # np.testing.assert_almost_equal(opt_func, -0.3857010, 4,
        #                                err_msg='calibrated NSE is not matching expected value')
        mohyse = MOHYSE()
        mohyse(
            TS,
            start_date=dt.datetime(1954, 1, 1),
            duration=208,
            area=4250.6,
            elevation=843.0,
            latitude=54.4848,
            longitude=-123.3659,
            params=model.calibrated_params,
        )

        np.testing.assert_almost_equal(
            mohyse.diagnostics["DIAG_NASH_SUTCLIFFE"], d["DIAG_NASH_SUTCLIFFE"], 4
        )


class TestHBVEC:
    def test_simple(self):
        model = HBVEC()
        params = (
            0.05984519,
            4.072232,
            2.001574,
            0.03473693,
            0.09985144,
            0.506052,
            3.438486,
            38.32455,
            0.4606565,
            0.06303738,
            2.277781,
            4.873686,
            0.5718813,
            0.04505643,
            0.877607,
            18.94145,
            2.036937,
            0.4452843,
            0.6771759,
            1.141608,
            1.024278,
        )

        model(
            TS,
            start_date=dt.datetime(2000, 1, 1),
            end_date=dt.datetime(2002, 1, 1),
            area=4250.6,
            elevation=843.0,
            latitude=54.4848,
            longitude=-123.3659,
            params=params,
            suppress_output=True,
        )

        d = model.diagnostics
        np.testing.assert_almost_equal(d["DIAG_NASH_SUTCLIFFE"], 0.0186633, 4)

    def test_evap(self):
        model = HBVEC()
        params = (
            0.05984519,
            4.072232,
            2.001574,
            0.03473693,
            0.09985144,
            0.506052,
            3.438486,
            38.32455,
            0.4606565,
            0.06303738,
            2.277781,
            4.873686,
            0.5718813,
            0.04505643,
            0.877607,
            18.94145,
            2.036937,
            0.4452843,
            0.6771759,
            1.141608,
            1.024278,
        )

        model(
            TS,
            start_date=dt.datetime(2000, 1, 1),
            end_date=dt.datetime(2002, 1, 1),
            area=4250.6,
            elevation=843.0,
            latitude=54.4848,
            longitude=-123.3659,
            params=params,
            suppress_output=True,
            evaporation="PET_OUDIN",
            ow_evaporation="PET_OUDIN",
        )


class TestHBVEC_OST:
    def test_simple(self):
        model = HBVEC_OST()

        model.configure(
            get_local_testdata("ostrich-gr4j-cemaneige/OstRandomNumbers.txt")
        )

        params = (
            0.05984519,
            4.072232,
            2.001574,
            0.03473693,
            0.09985144,
            0.506052,
            3.438486,
            38.32455,
            0.4606565,
            0.06303738,
            2.277781,
            4.873686,
            0.5718813,
            0.04505643,
            0.877607,
            18.94145,
            2.036937,
            0.4452843,
            0.6771759,
            1.141608,
            1.024278,
        )

        low = (
            -3.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.3,
            0.0,
            0.0,
            0.01,
            0.05,
            0.01,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.01,
            0.0,
            0.05,
            0.8,
            0.8,
        )
        high = (
            3.0,
            8.0,
            8.0,
            0.1,
            1.0,
            1.0,
            7.0,
            100.0,
            1.0,
            0.1,
            6.0,
            5.0,
            5.0,
            0.2,
            1.0,
            30.0,
            3.0,
            2.0,
            1.0,
            1.5,
            1.5,
        )

        model(
            TS,
            start_date=dt.datetime(1954, 1, 1),
            duration=208,
            area=4250.6,
            elevation=843.0,
            latitude=54.4848,
            longitude=-123.3659,
            params=params,
            lowerBounds=low,
            upperBounds=high,
            algorithm="DDS",
            random_seed=0,
            max_iterations=10,
        )

        d = model.diagnostics
        np.testing.assert_almost_equal(d["DIAG_NASH_SUTCLIFFE"], -2.25991e-01, 4)

        opt_para = astuple(model.calibrated_params)
        opt_func = model.obj_func

        # Random number seed: 123                         #
        # Budget:             10                          #      This is the setup used for testing:
        # Algorithm:          DDS                         #         shorter sim-period and lower budget
        # :StartDate          1954-01-01 00:00:00         #      First tested that example below matches
        # :Duration           208                         #
        np.testing.assert_almost_equal(
            opt_para,
            [
                -8.317931e-01,
                4.072232e00,
                2.001574e00,
                5.736299e-03,
                9.985144e-02,
                4.422529e-01,
                3.438486e00,
                8.055843e01,
                4.440133e-01,
                8.451082e-02,
                2.814201e00,
                7.327970e-01,
                1.119773e00,
                1.161223e-03,
                4.597179e-01,
                1.545857e01,
                1.223865e00,
                4.452843e-01,
                9.492006e-01,
                9.948123e-01,
                1.110682e00,
            ],
            4,
            err_msg="calibrated parameter set is not matching expected value",
        )
        np.testing.assert_almost_equal(
            opt_func,
            2.25991e-01,
            4,
            err_msg="calibrated NSE is not matching expected value",
        )

        # # Random number seed: 123                       #
        # # Budget:             50                        #      This is the setup in the Wiki:
        # # Algorithm:          DDS                       #      https://github.com/Ouranosinc/raven/wiki/
        # # :StartDate          1954-01-01 00:00:00       #      Technical-Notes#example-setups-for-environment-
        # # :Duration           20819                     #
        # np.testing.assert_almost_equal(opt_para, [5.984519E-02, 4.072232E+00, 2.001574E+00, 3.473693E-02,
        #                                           9.985144E-02, 5.060520E-01, 2.944343E+00, 3.832455E+01,
        #                                           4.606565E-01, 6.303738E-02, 2.277781E+00, 4.873686E+00,
        #                                           5.718813E-01, 4.505643E-02, 8.776511E-01, 1.894145E+01,
        #                                           2.036937E+00, 4.452843E-01, 6.771759E-01, 1.206053E+00,
        #                                           1.024278E+00], 4,
        #                                err_msg='calibrated parameter set is not matching expected value')
        # np.testing.assert_almost_equal(opt_func, -6.034670E-01, 4,
        #                                err_msg='calibrated NSE is not matching expected value')
        hbvec = HBVEC()
        hbvec(
            TS,
            start_date=dt.datetime(1954, 1, 1),
            duration=208,
            area=4250.6,
            elevation=843.0,
            latitude=54.4848,
            longitude=-123.3659,
            params=model.calibrated_params,
        )

        np.testing.assert_almost_equal(
            hbvec.diagnostics["DIAG_NASH_SUTCLIFFE"], d["DIAG_NASH_SUTCLIFFE"], 4
        )



# -*- coding: utf-8 -*-

# This information is located in its own file so that it can be loaded
# without importing the main package when its dependencies are not installed.
# See: https://packaging.python.org/guides/single-sourcing-package-version

__author__ = """David Huard"""
__email__ = "huard.david@ouranos.ca"
__version__ = "0.5.0"



from .blended import *
from .gr4jcn import *
from .hbvec import *
from .hmets import *
from .mohyse import *


def get_model(name):
    """Return the corresponding Raven emulated model instance.

    Parameters
    ----------
    name : str
      Model class name or model identifier.

    Returns
    -------
    Raven model instance
    """
    from ravenpy.models import emulators

    model_cls = getattr(emulators, name.upper(), None)

    if model_cls is None:
        raise ValueError("Model {} is not recognized.".format(name))

    return model_cls



import re

import netCDF4 as nc4
from click.testing import CliRunner

from ravenpy.cli import aggregate_forcings_to_hrus, generate_grid_weights
from ravenpy.config.commands import GridWeightsCommand
from ravenpy.utilities.testdata import get_local_testdata


class TestGenerateGridWeights:
    def test_generate_grid_weights_with_nc_input_and_2d_coords(self, tmp_path):
        runner = CliRunner()
        output_path = tmp_path / "bla.rvt"
        params = [
            get_local_testdata("raven-routing-sample/VIC_streaminputs.nc"),
            get_local_testdata("raven-routing-sample/finalcat_hru_info.zip"),
            "-o",
            output_path,
        ]
        params = map(str, params)

        result = runner.invoke(generate_grid_weights, params)

        assert result.exit_code == 0
        assert not result.exception

        output = output_path.read_text()

        assert ":NumberHRUs 51" in output
        assert ":NumberGridCells 100" in output
        assert len(output.split("\n")) == 216

        # check derived weights
        weight = float(re.search("1 52 (.+)", output).group(1))
        assert abs(weight - 0.2610203097218425) < 1e-04

    def test_generate_grid_weights_with_multiple_subids(self, tmp_path):
        # currently exactly same output as "test_generate_grid_weights_with_nc_input_and_2d_coords"
        # needs a "routing-file-path" with multiple gauges
        runner = CliRunner()
        output_path = tmp_path / "bla.rvt"
        params = [
            get_local_testdata("raven-routing-sample/VIC_streaminputs.nc"),
            get_local_testdata("raven-routing-sample/finalcat_hru_info.zip"),
            "-s",
            "7202",
            "-s",
            "6248",
            "-o",
            output_path,
        ]
        params = map(str, params)

        result = runner.invoke(generate_grid_weights, params)

        assert result.exit_code == 0
        assert not result.exception

        output = output_path.read_text()

        assert ":NumberHRUs 51" in output
        assert ":NumberGridCells 100" in output
        assert len(output.split("\n")) == 216

        # check derived weights
        weight = float(re.search("1 52 (.+)", output).group(1))
        assert abs(weight - 0.2610203097218425) < 1e-04

    def test_generate_grid_weights_with_nc_input_and_1d_coords(self, tmp_path):
        runner = CliRunner()
        output_path = tmp_path / "bla.rvt"
        params = [
            get_local_testdata("raven-routing-sample/era5-test-dataset-crop.nc"),
            get_local_testdata("raven-routing-sample/finalcat_hru_info.zip"),
            "--var-names",
            "longitude",
            "latitude",
            "-o",
            output_path,
        ]
        params = map(str, params)

        result = runner.invoke(generate_grid_weights, params)

        assert result.exit_code == 0
        assert not result.exception

        output = output_path.read_text()

        assert ":NumberHRUs 51" in output
        assert ":NumberGridCells 9801" in output
        assert len(output.split("\n")) == 128

        # check derived weights
        weight = float(re.search("4 3731 (.+)", output).group(1))
        assert abs(weight - 0.0034512752779023515) < 1e-04

    def test_generate_grid_weights_with_shp_input(self, tmp_path):
        runner = CliRunner()
        output_path = tmp_path / "bla.rvt"
        params = [
            get_local_testdata("raven-routing-sample/OTT_sub.zip"),
            get_local_testdata("raven-routing-sample/finalcat_hru_info.zip"),
            "-o",
            output_path,
        ]
        params = map(str, params)

        result = runner.invoke(generate_grid_weights, params)

        assert result.exit_code == 0
        assert not result.exception

        output = output_path.read_text()

        assert ":NumberHRUs 51" in output
        assert ":NumberGridCells 810" in output
        assert len(output.split("\n")) == 230

        # check derived weights
        weight = float(re.search("13 238 (.+)", output).group(1))
        assert abs(weight - 0.5761414847779369) < 1e-04

    def test_generate_grid_weights_with_weight_rescaling(self, tmp_path):
        runner = CliRunner()
        output_path = tmp_path / "bla.rvt"
        params = [
            get_local_testdata("raven-routing-sample/OTT_sub.zip"),
            get_local_testdata("raven-routing-sample/finalcat_hru_info.zip"),
            "--area-error-threshold",
            "0.42",
            "-o",
            output_path,
        ]
        params = map(str, params)

        result = runner.invoke(generate_grid_weights, params)

        assert result.exit_code == 0
        assert not result.exception

        output = output_path.read_text()

        assert ":NumberHRUs 51" in output
        assert ":NumberGridCells 810" in output
        assert len(output.split("\n")) == 230

        # check derived weights
        weight = float(re.search("13 238 (.+)", output).group(1))
        assert abs(weight - 0.9851111335377887) < 1e-04


class TestAggregateForcingsToHRUs:
    def test_aggregate_forcings_to_hrus(self, tmp_path):
        runner = CliRunner()
        output_nc_file_path = tmp_path / "aggreg.nc"
        output_weight_file_path = tmp_path / "weight_aggreg.rvt"
        params = [
            get_local_testdata("raven-routing-sample/VIC_streaminputs.nc"),
            get_local_testdata("raven-routing-sample/VIC_streaminputs_weights.rvt"),
            "-v",
            "Streaminputs",
            "--output-nc-file",
            output_nc_file_path,
            "--output-weight-file",
            output_weight_file_path,
        ]
        params = map(str, params)

        result = runner.invoke(aggregate_forcings_to_hrus, params)

        assert result.exit_code == 0
        assert not result.exception

        output_rvt = output_weight_file_path.read_text()

        gws = GridWeightsCommand.parse(output_rvt)

        new_weights = gws.data

        # check new weights
        assert new_weights[0][0] == 1  # These are the HRU-IDs

        assert new_weights[0][1] == 0  # These need to be exactly [0,1,2,3,...,nHRU]
        assert new_weights[1][1] == 1  # These need to be exactly [0,1,2,3,...,nHRU]
        assert new_weights[2][1] == 2  # These need to be exactly [0,1,2,3,...,nHRU]
        assert new_weights[3][1] == 3  # These need to be exactly [0,1,2,3,...,nHRU]

        assert new_weights[0][2] == 1.0  # All new_weights[:][2] need to be 1.0
        assert new_weights[1][2] == 1.0  # All new_weights[:][2] need to be 1.0
        assert new_weights[2][2] == 1.0  # All new_weights[:][2] need to be 1.0
        assert new_weights[3][2] == 1.0  # All new_weights[:][2] need to be 1.0

        # check aggregated NetCDF file
        nc_in = nc4.Dataset(output_nc_file_path, "r")
        val = nc_in.variables["Streaminputs"][:]
        nc_in.close()

        assert abs(val[0, 0] - 0.017309) < 1e-04
        assert abs(val[16071, 0] - 0.569977) < 1e-04
        assert abs(val[0, 50] - 0.010276) < 1e-04
        assert abs(val[16071, 50] - 0.516639) < 1e-04

    def test_aggregate_forcings_to_hrus_with_nodata(self, tmp_path):
        runner = CliRunner()
        output_nc_file_path = tmp_path / "aggreg.nc"
        output_weight_file_path = tmp_path / "weight_aggreg.rvt"
        params = [
            get_local_testdata("raven-routing-sample/VIC_test_nodata.nc"),
            get_local_testdata("raven-routing-sample/VIC_test_nodata_weights.rvt"),
            "-v",
            "et",
            "--dim-names",
            "rlon",
            "rlat",
            "--output-nc-file",
            output_nc_file_path,
            "--output-weight-file",
            output_weight_file_path,
        ]
        params = map(str, params)

        result = runner.invoke(aggregate_forcings_to_hrus, params)

        assert result.exit_code == 0
        assert not result.exception

        output_rvt = output_weight_file_path.read_text()

        gws = GridWeightsCommand.parse(output_rvt)

        new_weights = gws.data

        # check new weights
        assert new_weights[0][0] == 1  # These are the HRU-IDs
        assert new_weights[1][0] == 2  # These are the HRU-IDs
        assert new_weights[2][0] == 3  # These are the HRU-IDs

        assert new_weights[0][1] == 0  # These need to be exactly [0,1,2,3,...,nHRU]
        assert new_weights[1][1] == 1  # These need to be exactly [0,1,2,3,...,nHRU]
        assert new_weights[2][1] == 2  # These need to be exactly [0,1,2,3,...,nHRU]

        assert new_weights[0][2] == 1.0  # All new_weights[:][2] need to be 1.0
        assert new_weights[1][2] == 1.0  # All new_weights[:][2] need to be 1.0
        assert new_weights[2][2] == 1.0  # All new_weights[:][2] need to be 1.0

        # check aggregated NetCDF file
        nc_in = nc4.Dataset(output_nc_file_path, "r")
        val = nc_in.variables["et"][:]
        nc_in.close()

        # aggregated time series for HRU #1
        # HRU #1 = [ 10% cell 0 ; 20% cell 1 ; 30% cell 3 ; 40% cell 4 ]
        # (cell 4 is NODATA for each time step; cell 3 is NODATA at 3rd time step)
        assert (
            abs(val[0, 0] - 2.8333333) < 1e-04
        )  # = 0.1*1 + 0.2*2 + 0.3*4 + 0.4*NODATA
        #                                           # = 0.1/0.6*1 + 0.2/0.6*2 + 0.3/0.6*4
        assert (
            abs(val[1, 0] - 3.0000000) < 1e-04
        )  # = 0.1*3 + 0.2*3 + 0.3*3 + 0.4*NODATA
        #                                           # = 0.1/0.6*3 + 0.2/0.6*3 + 0.3/0.6*3
        assert (
            abs(val[2, 0] - 7.3333333) < 1e-04
        )  # = 0.1*4 + 0.2*9 + 0.3*NODATA + 0.4*NODATA
        #                                           # = 0.1/0.3*4 + 0.2/0.3*9
        assert (
            abs(val[3, 0] - 2.8333333) < 1e-04
        )  # = 0.1*1 + 0.2*2 + 0.3*4 + 0.4*NODATA
        #                                           # = 0.1/0.6*1 + 0.2/0.6*2 + 0.3/0.6*4

        # aggregated time series for HRU #2
        # HRU #2 = [ 20% cell 1 ; 80% cell 2 ]
        # (no cell ever contains NODATA)
        assert abs(val[0, 1] - 2.8) < 1e-04  # = 0.2*2 + 0.8*3
        assert abs(val[1, 1] - 3.0) < 1e-04  # = 0.2*3 + 0.8*3
        assert abs(val[2, 1] - 6.6) < 1e-04  # = 0.2*9 + 0.8*6
        assert abs(val[3, 1] - 2.8) < 1e-04  # = 0.2*2 + 0.8*3

        # aggregated time series for HRU #3
        # HRU #3 = [ 20% cell 1 ; 10% cell 2 ; 70% cell 5 ]
        # (cell 5 is NODATA at 3rd time step)
        assert abs(val[0, 2] - 4.9) < 1e-04  # = 0.2*2 + 0.1*3 + 0.7*6
        assert abs(val[1, 2] - 3.0) < 1e-04  # = 0.2*3 + 0.1*3 + 0.7*3
        assert abs(val[2, 2] - 8.0) < 1e-04  # = 0.2*9 + 0.1*6 + 0.7*NODATA
        #                                    # = 0.2/(0.2+0.1)*9 + 0.1/(0.2+0.1)*6
        assert abs(val[3, 2] - 4.9) < 1e-04  # = 0.2*2 + 0.1*3 + 0.7*6

        # aggregated time series for HRU #4
        # HRU #3 = [ 40% cell 3 ; 60% cell 4 ]
        # (cell3 is NODATA at 3rd time step; cell 4 is NODATA for each time step)
        assert abs(val[0, 3] - 4.0) < 1e-04  # = 0.4*4 + 0.6*NODATA = 1.0*4
        assert abs(val[1, 3] - 3.0) < 1e-04  # = 0.4*3 + 0.6*NODATA = 1.0*3
        assert val[2, 3].mask  # = 0.4*NODATA + 0.6*NODATA
        assert abs(val[2, 3].data - 0.0) < 1e-04  # = 0.4*NODATA + 0.6*NODATA
        assert abs(val[3, 3] - 4.0) < 1e-04  # = 0.4*4 + 0.6*NODATA = 1.0*4



import click

from ravenpy import __version__

from .aggregate_forcings_to_hrus import aggregate_forcings_to_hrus
from .generate_grid_weights import generate_grid_weights


@click.group()
@click.version_option(__version__)
def cli():
    pass


def main():
    cli.add_command(generate_grid_weights)
    cli.add_command(aggregate_forcings_to_hrus)
    cli()



from collections import defaultdict
from pathlib import Path

from pydantic.dataclasses import dataclass

from ravenpy.config.commands import HRU, LU, BasinIndexCommand, HRUState, Sub
from ravenpy.config.rvs import RVI
from ravenpy.models.base import Ostrich, Raven

from .gr4jcn import GR4JCN


class BLENDED(Raven):
    @dataclass
    class Params:
        par_x01: float = None
        par_x02: float = None
        par_x03: float = None
        par_x04: float = None
        par_x05: float = None
        par_x06: float = None
        par_x07: float = None
        par_x08: float = None
        par_x09: float = None
        par_x10: float = None
        par_x11: float = None
        par_x12: float = None
        par_x13: float = None
        par_x14: float = None
        par_x15: float = None
        par_x16: float = None
        par_x17: float = None
        par_x18: float = None
        par_x19: float = None
        par_x20: float = None
        par_x21: float = None
        par_x22: float = None
        par_x23: float = None
        par_x24: float = None
        par_x25: float = None
        par_x26: float = None
        par_x27: float = None
        par_x28: float = None
        par_x29: float = None
        par_x30: float = None
        par_x31: float = None
        par_x32: float = None
        par_x33: float = None
        par_x34: float = None
        par_x35: float = None
        par_r01: float = None
        par_r02: float = None
        par_r03: float = None
        par_r04: float = None
        par_r05: float = None
        par_r06: float = None
        par_r07: float = None
        par_r08: float = None

    @dataclass
    class DerivedParams:
        POW_X04: float = None
        POW_X11: float = None
        SUM_X09_X10: float = None
        SUM_X13_X14: float = None
        SUM_X24_X25: float = None

    @dataclass
    class ForestHRU(HRU):
        land_use_class: str = "FOREST"
        veg_class: str = "FOREST"
        soil_profile: str = "DEFAULT_P"
        aquifer_profile: str = "[NONE]"
        terrain_class: str = "DEFAULT_T"

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)

        self.config.update(
            identifier="blended",
            hrus=(BLENDED.ForestHRU(),),
            subbasins=(
                Sub(
                    subbasin_id=1,
                    name="sub_001",
                    downstream_id=-1,
                    profile="None",
                    gauged=True,
                ),
            ),
            params=BLENDED.Params(),
            derived_params=BLENDED.DerivedParams(),
            land_use_classes=(
                LU("FOREST", impermeable_frac=0.0, forest_coverage=0.02345),
            ),
            evaporation="PET_OUDIN",
            rain_snow_fraction=RVI.RainSnowFractionOptions.HBV,
        )

        #########
        # R V P #
        #########

        rvp_tmpl = """
        # tied parameters:
        # (it is important for OSTRICH to find every parameter place holder somewhere in this file)
        # (without this "5.421821E-02" and "1.596675E-01" and "2.169724E-01" wouldn't be detectable)
        #    para_1.705806E+00 = 1.705806E+00 =  para_x24 + para_x25 = 1.651588E+00 + 5.421821E-02
        #    para_2.070342E-01 = 2.070342E-01 =  para_x13 + para_x14 = 4.736668E-02 + 1.596675E-01
        #    para_2.518350E-01 = 2.518350E-01 =  para_x09 + para_x10 = 3.486263E-02 + 2.169724E-01
        #    para_2.254976E-04     = 2.254976E-04     =  10^(para_x04)       = 10^-3.646858E+00
        #    para_2.279250E-04     = 2.279250E-04     =  10^(para_x11)       = 10^-3.642208E+00

        #-----------------------------------------------------------------
        # Soil Classes
        #-----------------------------------------------------------------
        :SoilClasses
          :Attributes,
          :Units,
          TOPSOIL,
          PHREATIC,
          DEEP_GW
        :EndSoilClasses

        #-----------------------------------------------------------------
        # Land Use Classes
        #-----------------------------------------------------------------
        {land_use_classes}

        #-----------------------------------------------------------------
        # Vegetation Classes
        #-----------------------------------------------------------------
        :VegetationClasses,
          :Attributes,        MAX_HT,       MAX_LAI, MAX_LEAF_COND,
               :Units,             m,          none,      mm_per_s,
               FOREST,             4,             5,             5,
        :EndVegetationClasses

        #-----------------------------------------------------------------
        # Soil Profiles
        #-----------------------------------------------------------------
        :SoilProfiles
                 LAKE, 0
                 ROCK, 0
                 DEFAULT_P, 3, TOPSOIL,  {params.par_x29}, PHREATIC, {params.par_x30}, DEEP_GW, 1e6
        #        DEFAULT_P, 3, TOPSOIL,         x(29), PHREATIC,        x(30), DEEP_GW, 1e6
        :EndSoilProfiles

        #-----------------------------------------------------------------
        # Terrain Classes
        #-----------------------------------------------------------------
        :TerrainClasses
          :Attributes,        hillslope_len, drainage_dens,                 lambda,
               :Units,                   ??,            ??,                     ??
            DEFAULT_T,                  1.0,           1.0,           {params.par_x07}
        #                                                     TOPMODEL_LAMBDA x(7)
        :EndTerrainClasses

        #-----------------------------------------------------------------
        # Global Parameters
        #-----------------------------------------------------------------
        :GlobalParameter         SNOW_SWI_MIN {params.par_x13}  #  x(13)
        :GlobalParameter         SNOW_SWI_MAX {derived_params.SUM_X13_X14}     #  x(13)+x(14)
        :GlobalParameter     SWI_REDUCT_COEFF {params.par_x15}  #  x(15)
        :GlobalParameter             SNOW_SWI {params.par_x19}  #  x(19)
        :GlobalParameter        RAINSNOW_TEMP {params.par_x31}  #  x(31)
        :GlobalParameter       RAINSNOW_DELTA {params.par_x32}  #  x(32)
        #:GlobalParameter      TOC_MULTIPLIER 1.0               #


        #-----------------------------------------------------------------
        # Soil Parameters
        #-----------------------------------------------------------------
        :SoilParameterList
          :Parameters,        POROSITY,           PERC_COEFF,  PET_CORRECTION,  BASEFLOW_COEFF,        B_EXP,     HBV_BETA, MAX_BASEFLOW_RATE,   BASEFLOW_N, FIELD_CAPACITY,     SAT_WILT,
               :Units,               -,                  1/d,               -,             1/d
              TOPSOIL,             1.0,         {params.par_x28},    {params.par_x08},       {derived_params.POW_X04}, {params.par_x02}, {params.par_x03}, 	 {params.par_x05}, {params.par_x06},  {derived_params.SUM_X09_X10}, {params.par_x09},
             PHREATIC,             1.0,         {params.par_x35},             0.0,       {derived_params.POW_X11},          0.0,          0.0, 	          0.0, {params.par_x12},            0.0,          0.0,
              DEEP_GW,             1.0,             0.0,                  0.0,             0.0,          0.0,          0.0, 	          0.0,          0.0,            0.0,          0.0,
         #    TOPSOIL,             1.0,           x(28),                x(08),           x(04),        x(02),        x(03), 	        x(05),        x(06),    x(09)+x(10),        x(09),
         #   PHREATIC,             1.0,           x(35),                  0.0,           x(11),          0.0,          0.0, 	          0.0,        x(12),            0.0,          0.0,
         #    DEEP_GW,             1.0,             0.0,                  0.0,             0.0,          0.0,          0.0, 	          0.0,          0.0,            0.0,          0.0,
        :EndSoilParameterList

        #-----------------------------------------------------------------
        # Land Use Parameters
        #-----------------------------------------------------------------
        :LandUseParameterList
          :Parameters, MIN_MELT_FACTOR,     MAX_MELT_FACTOR,    DD_MELT_TEMP,  DD_AGGRADATION, REFREEZE_FACTOR, REFREEZE_EXP, DD_REFREEZE_TEMP, HMETS_RUNOFF_COEFF,
               :Units,          mm/d/C,              mm/d/C,               C,            1/mm,          mm/d/C,            -,                C,                  -,
            [DEFAULT],    {params.par_x24},       {derived_params.SUM_X24_X25},    {params.par_x26},    {params.par_x27},    {params.par_x18}, {params.par_x17},     {params.par_x16},       {params.par_x01},
        #                        x(24),         x(24)+x(25),           x(26),           x(27),           x(18),        x(17),            x(16),              x(01),
        :EndLandUseParameterList
        :LandUseParameterList
          :Parameters,   GAMMA_SHAPE,     GAMMA_SCALE,    GAMMA_SHAPE2,    GAMMA_SCALE2,    FOREST_SPARSENESS,
               :Units,             -,               -,               -,               -,                    -,
            [DEFAULT],  {params.par_x20},    {params.par_x21},    {params.par_x22},    {params.par_x23},                  0.0,
            #                  x(20),           x(21),           x(22),           x(23),                  0.0,
        :EndLandUseParameterList

        #-----------------------------------------------------------------
        # Vegetation Parameters
        #-----------------------------------------------------------------
        :VegetationParameterList
          :Parameters,  RAIN_ICEPT_PCT,  SNOW_ICEPT_PCT,    SAI_HT_RATIO
               :Units,               -,               -,               -
            [DEFAULT],             0.0,             0.0,             0.0
        :EndVegetationParameterList

        :SeasonalRelativeLAI
          FOREST, 1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0
        :EndSeasonalRelativeLAI
        :SeasonalRelativeHeight
          FOREST, 1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0
        :EndSeasonalRelativeHeight
        """
        self.config.rvp.set_tmpl(rvp_tmpl)

        #########
        # R V I #
        #########

        rvi_tmpl = """
        :PotentialMeltMethod     POTMELT_HMETS
        :RainSnowFraction        {rain_snow_fraction}
        :Evaporation             {evaporation}         # PET_OUDIN
        :CatchmentRoute          ROUTE_DUMP
        :Routing                 ROUTE_NONE
        :SoilModel               SOIL_MULTILAYER 3

        :Alias DELAYED_RUNOFF CONVOLUTION[1]

        :HydrologicProcesses
          :Precipitation   RAVEN_DEFAULT                         ATMOS_PRECIP   MULTIPLE
          :ProcessGroup #infiltration group
                        :Infiltration    INF_HMETS               PONDED_WATER   MULTIPLE
                        :Infiltration    INF_VIC_ARNO            PONDED_WATER   MULTIPLE
                        :Infiltration    INF_HBV                 PONDED_WATER   MULTIPLE
          :EndProcessGroup CALCULATE_WTS {params.par_r01} {params.par_r02}
          #                              para_r01 para_r02
                          :Overflow      OVERFLOW_RAVEN          SOIL[0]        DELAYED_RUNOFF
          :ProcessGroup #quickflow group
                        :Baseflow        BASE_LINEAR_ANALYTIC    SOIL[0]        SURFACE_WATER   # interflow, really
                        :Baseflow        BASE_VIC                SOIL[0]        SURFACE_WATER
                        :Baseflow        BASE_TOPMODEL           SOIL[0]        SURFACE_WATER
          :EndProcessGroup CALCULATE_WTS {params.par_r03} {params.par_r04}
          #                              para_r03 para_r04
          :Percolation                   PERC_LINEAR             SOIL[0]        SOIL[1]         # recharge
            :Overflow                    OVERFLOW_RAVEN          SOIL[1]        DELAYED_RUNOFF
          :Percolation                   PERC_LINEAR             SOIL[1]        SOIL[2]         # loss to deep groundwater (simplifies to HMETS when PERC_COEFF DEEP_GW=0)
          :ProcessGroup #evaporation group
                        :SoilEvaporation SOILEVAP_ALL            SOIL[0]        ATMOSPHERE      # AET
                        :SoilEvaporation SOILEVAP_TOPMODEL       SOIL[0]        ATMOSPHERE      # AET
          :EndProcessGroup CALCULATE_WTS {params.par_r05}
          #                              para_r05
          :Convolve                      CONVOL_GAMMA            CONVOLUTION[0] SURFACE_WATER   # 'surface runoff'
          :Convolve                      CONVOL_GAMMA_2          DELAYED_RUNOFF SURFACE_WATER   # 'delayed runoff'
          :ProcessGroup #quickflow group
                        :Baseflow        BASE_LINEAR_ANALYTIC    SOIL[1]        SURFACE_WATER
                        :Baseflow        BASE_POWER_LAW          SOIL[1]        SURFACE_WATER
          :EndProcessGroup CALCULATE_WTS {params.par_r06}
          #                              para_r06
          :ProcessGroup #snow balance group
                        :SnowBalance     SNOBAL_HMETS            MULTIPLE       MULTIPLE
                        :SnowBalance     SNOBAL_SIMPLE_MELT      SNOW           PONDED_WATER
                        :SnowBalance     SNOBAL_HBV              MULTIPLE       MULTIPLE
                        #:SnowBalance    SNOBAL_GAWSER           MULTIPLE       MULTIPLE
          :EndProcessGroup CALCULATE_WTS {params.par_r07} {params.par_r08}
          #                              para_r07 para_r08
        :EndHydrologicProcesses
        """
        self.config.rvi.set_tmpl(rvi_tmpl)

        self.config.rvi.params = self.config.rvp.params

    def derived_parameters(self):
        self.config.rvp.derived_params.SUM_X09_X10 = (
            self.config.rvp.params.par_x10
        )  # + self.config.rvp.params.par_x09
        self.config.rvp.derived_params.SUM_X13_X14 = (
            self.config.rvp.params.par_x14
        )  # + self.config.rvp.params.par_x13
        self.config.rvp.derived_params.SUM_X24_X25 = (
            self.config.rvp.params.par_x25
        )  # + self.config.rvp.params.par_x24
        # 10.0**self.config.rvp.params.par_x04  #
        self.config.rvp.derived_params.POW_X04 = self.config.rvp.params.par_x04
        # 10.0**self.config.rvp.params.par_x11  #
        self.config.rvp.derived_params.POW_X11 = self.config.rvp.params.par_x11

        topsoil_hlf = self.config.rvp.params.par_x29 * 0.5 * 1000
        phreatic_hlf = self.config.rvp.params.par_x30 * 0.5 * 1000
        hru_state = HRUState(soil0=topsoil_hlf, soil1=phreatic_hlf)
        self.config.rvc.set_hru_state(hru_state)

        self.config.rvt.rain_correction = self.config.rvp.params.par_x33
        self.config.rvt.snow_correction = self.config.rvp.params.par_x34


class BLENDED_OST(Ostrich, BLENDED):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)

        self.config.update(
            identifier="blended-ost",
            algorithm="DDS",
            max_iterations=50,
            lowerBounds=BLENDED.Params(),
            upperBounds=BLENDED.Params(),
            suppress_output=True,
        )

        ####################
        # R V C (OST TMPL) #
        ####################

        rvc_tmpl = """
        # tied parameters:
        # (it is important for OSTRICH to find every parameter place holder somewhere in this file)
        # (without this "par_x29" and "par_x30" wouldn't be detectable)
        #    para_half_x29 = para_x29 * 1000. / 2. = par_x29 / 2. [m] = half_x29 [mm]
        #    para_half_x30 = para_x30 * 1000. / 2. = par_x30 / 2. [m] = half_x30 [mm]

        # initialize to 1/2 full
        #:UniformInitialConditions SOIL[0] half_x29 # x(29)*1000/2 [mm]
        #:UniformInitialConditions SOIL[1] half_x30 # x(30)*1000/2 [mm]

        :HRUStateVariableTable (formerly :IntialConditionsTable)
           :Attributes SOIL[0] SOIL[1]
           :Units mm mm
           1 half_x29 half_x30
        :EndHRUStateVariableTable
        """
        self.config.rvc.set_tmpl(rvc_tmpl, is_ostrich=True)

        ####################
        # R V I (OST TMPL) #
        ####################

        rvi_tmpl = """
        :Calendar              {calendar}
        :RunName               {run_name}-{run_index}
        :StartDate             {start_date}
        :EndDate               {end_date}
        :TimeStep              {time_step}
        :Method                ORDERED_SERIES

        :PotentialMeltMethod     POTMELT_HMETS
        :RainSnowFraction        {rain_snow_fraction}
        :Evaporation             {evaporation}         # PET_OUDIN
        :CatchmentRoute          ROUTE_DUMP
        :Routing                 ROUTE_NONE
        :SoilModel               SOIL_MULTILAYER 3

        :Alias DELAYED_RUNOFF CONVOLUTION[1]

        :HydrologicProcesses
          :Precipitation   RAVEN_DEFAULT                         ATMOS_PRECIP   MULTIPLE
          :ProcessGroup #infiltration group
                        :Infiltration    INF_HMETS               PONDED_WATER   MULTIPLE
                        :Infiltration    INF_VIC_ARNO            PONDED_WATER   MULTIPLE
                        :Infiltration    INF_HBV                 PONDED_WATER   MULTIPLE
          :EndProcessGroup CALCULATE_WTS par_r01 par_r02
                          :Overflow      OVERFLOW_RAVEN          SOIL[0]        DELAYED_RUNOFF
          :ProcessGroup #quickflow group
                        :Baseflow        BASE_LINEAR_ANALYTIC    SOIL[0]        SURFACE_WATER   # interflow, really
                        :Baseflow        BASE_VIC                SOIL[0]        SURFACE_WATER
                        :Baseflow        BASE_TOPMODEL           SOIL[0]        SURFACE_WATER
          :EndProcessGroup CALCULATE_WTS par_r03 par_r04
          :Percolation                   PERC_LINEAR             SOIL[0]        SOIL[1]         # recharge
            :Overflow                    OVERFLOW_RAVEN          SOIL[1]        DELAYED_RUNOFF
          :Percolation                   PERC_LINEAR             SOIL[1]        SOIL[2]         # loss to deep groundwater (simplifies to HMETS when PERC_COEFF DEEP_GW=0)
          :ProcessGroup #evaporation group
                        :SoilEvaporation SOILEVAP_ALL            SOIL[0]        ATMOSPHERE      # AET
                        :SoilEvaporation SOILEVAP_TOPMODEL       SOIL[0]        ATMOSPHERE      # AET
          :EndProcessGroup CALCULATE_WTS par_r05
          :Convolve                      CONVOL_GAMMA            CONVOLUTION[0] SURFACE_WATER   # 'surface runoff'
          :Convolve                      CONVOL_GAMMA_2          DELAYED_RUNOFF SURFACE_WATER   # 'delayed runoff'
          :ProcessGroup #quickflow group
                        :Baseflow        BASE_LINEAR_ANALYTIC    SOIL[1]        SURFACE_WATER
                        :Baseflow        BASE_POWER_LAW          SOIL[1]        SURFACE_WATER
          :EndProcessGroup CALCULATE_WTS par_r06
          :ProcessGroup #snow balance group
                        :SnowBalance     SNOBAL_HMETS            MULTIPLE       MULTIPLE
                        :SnowBalance     SNOBAL_SIMPLE_MELT      SNOW           PONDED_WATER
                        :SnowBalance     SNOBAL_HBV              MULTIPLE       MULTIPLE
                        #:SnowBalance    SNOBAL_GAWSER           MULTIPLE       MULTIPLE
          :EndProcessGroup CALCULATE_WTS par_r07 par_r08
        :EndHydrologicProcesses

        #:CreateRVPTemplate

        #---------------------------------------------------------
        # Output Options
        #
        :EvaluationMetrics NASH_SUTCLIFFE RMSE KLING_GUPTA
        :SuppressOutput
        :SilentMode
        :DontWriteWatershedStorage
        """
        self.config.rvi.set_tmpl(rvi_tmpl, is_ostrich=True)

        ####################
        # R V P (OST TMPL) #
        ####################

        rvp_tmpl = """
        # tied parameters:
        # (it is important for OSTRICH to find every parameter place holder somewhere in this file)
        # (without this "par_x25" and "par_x14" and "par_x10" wouldn't be detectable)
        #    para_sum_x24_x25 = sum_x24_x25 =  para_x24 + para_x25 = par_x24 + par_x25
        #    para_sum_x13_x14 = sum_x13_x14 =  para_x13 + para_x14 = par_x13 + par_x14
        #    para_sum_x09_x10 = sum_x09_x10 =  para_x09 + para_x10 = par_x09 + par_x10
        #    para_pow_x04     = pow_x04     =  10^(para_x04)       = 10^par_x04
        #    para_pow_x11     = pow_x11     =  10^(para_x11)       = 10^par_x11

        # RVT should contain
        #  :RainCorrection par_x33
        #  :SnowCorrection par_x34

        #-----------------------------------------------------------------
        # Soil Classes
        #-----------------------------------------------------------------
        :SoilClasses
          :Attributes,
          :Units,
          TOPSOIL,
          PHREATIC,
          DEEP_GW
        :EndSoilClasses

        #-----------------------------------------------------------------
        # Land Use Classes
        #-----------------------------------------------------------------
        {land_use_classes}

        #-----------------------------------------------------------------
        # Vegetation Classes
        #-----------------------------------------------------------------
        :VegetationClasses,
          :Attributes,        MAX_HT,       MAX_LAI, MAX_LEAF_COND,
               :Units,             m,          none,      mm_per_s,
               FOREST,             4,             5,             5,
        :EndVegetationClasses

        #-----------------------------------------------------------------
        # Soil Profiles
        #-----------------------------------------------------------------
        :SoilProfiles
                 LAKE, 0
                 ROCK, 0
          DEFAULT_P, 3, TOPSOIL,    par_x29, PHREATIC,    par_x30, DEEP_GW, 1e6
        # DEFAULT_P, 3, TOPSOIL,      x(29), PHREATIC,      x(30), DEEP_GW, 1e6
        :EndSoilProfiles

        #-----------------------------------------------------------------
        # Terrain Classes
        #-----------------------------------------------------------------
        :TerrainClasses
          :Attributes,        hillslope_len, drainage_dens,            lambda,
               :Units,                   ??,            ??,                ??
            DEFAULT_T,                  1.0,           1.0,           par_x07
        #                                                     TOPMODEL_LAMBDA x(7)
        :EndTerrainClasses

        #-----------------------------------------------------------------
        # Global Parameters
        #-----------------------------------------------------------------
        :GlobalParameter         SNOW_SWI_MIN par_x13            # x(13)
        :GlobalParameter         SNOW_SWI_MAX sum_x13_x14   # x(13)+x(14)
        :GlobalParameter     SWI_REDUCT_COEFF par_x15            # x(15)
        :GlobalParameter             SNOW_SWI par_x19            # x(19)
        :GlobalParameter        RAINSNOW_TEMP par_x31            # x(31)
        :GlobalParameter       RAINSNOW_DELTA par_x32            # x(32)
        #:GlobalParameter      TOC_MULTIPLIER 1.0                   #


        #-----------------------------------------------------------------
        # Soil Parameters
        #-----------------------------------------------------------------
        :SoilParameterList
          :Parameters,        POROSITY,      PERC_COEFF,  PET_CORRECTION,  BASEFLOW_COEFF,      B_EXP,   HBV_BETA, MAX_BASEFLOW_RATE, BASEFLOW_N,      FIELD_CAPACITY,   SAT_WILT,
               :Units,               -,             1/d,               -,             1/d
              TOPSOIL,             1.0,         par_x28,         par_x08,         pow_x04,    par_x02,    par_x03,           par_x05,    par_x06,         sum_x09_x10,    par_x09,
             PHREATIC,             1.0,         par_x35,             0.0,         pow_x11,        0.0,        0.0,               0.0,    par_x12,                 0.0,        0.0,
              DEEP_GW,             1.0,             0.0,             0.0,             0.0,        0.0,        0.0,               0.0,        0.0,                 0.0,        0.0,
         #    TOPSOIL,             1.0,           x(28),           x(08),           x(04),      x(02),      x(03),             x(05),      x(06),         x(09)+x(10),      x(09),
         #   PHREATIC,             1.0,           x(35),             0.0,           x(11),        0.0,        0.0,               0.0,      x(12),                 0.0,        0.0,
         #    DEEP_GW,             1.0,             0.0,             0.0,             0.0,        0.0,        0.0,               0.0,        0.0,                 0.0,        0.0,
        :EndSoilParameterList

        #-----------------------------------------------------------------
        # Land Use Parameters
        #-----------------------------------------------------------------
        :LandUseParameterList
          :Parameters, MIN_MELT_FACTOR,     MAX_MELT_FACTOR,    DD_MELT_TEMP,  DD_AGGRADATION, REFREEZE_FACTOR, REFREEZE_EXP, DD_REFREEZE_TEMP, HMETS_RUNOFF_COEFF,
               :Units,          mm/d/C,              mm/d/C,               C,            1/mm,          mm/d/C,            -,                C,                  -,
            [DEFAULT],         par_x24,         sum_x24_x25,         par_x26,         par_x27,         par_x18,      par_x17,          par_x16,            par_x01,
        #                        x(24),         x(24)+x(25),           x(26),           x(27),           x(18),        x(17),            x(16),              x(01),
        :EndLandUseParameterList
        :LandUseParameterList
          :Parameters,   GAMMA_SHAPE,     GAMMA_SCALE,    GAMMA_SHAPE2,    GAMMA_SCALE2,    FOREST_SPARSENESS,
               :Units,             -,               -,               -,               -,                    -,
            [DEFAULT],       par_x20,         par_x21,         par_x22,         par_x23,                  0.0,
            #                  x(20),           x(21),           x(22),           x(23),                  0.0,
        :EndLandUseParameterList

        #-----------------------------------------------------------------
        # Vegetation Parameters
        #-----------------------------------------------------------------
        :VegetationParameterList
          :Parameters,  RAIN_ICEPT_PCT,  SNOW_ICEPT_PCT,    SAI_HT_RATIO
               :Units,               -,               -,               -
            [DEFAULT],             0.0,             0.0,             0.0
        :EndVegetationParameterList

        :SeasonalRelativeLAI
          FOREST, 1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0
        :EndSeasonalRelativeLAI
        :SeasonalRelativeHeight
          FOREST, 1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0
        :EndSeasonalRelativeHeight
        """
        self.config.rvp.set_tmpl(rvp_tmpl, is_ostrich=True)

        ####################
        # R V T (OST TMPL) #
        ####################

        rvt_tmpl = """
        {gauge}

        {forcing_list}

        {observed_data}
        """
        self.config.rvt.set_tmpl(rvt_tmpl, is_ostrich=True)

        ##########
        # O S T  #
        ##########

        ost_tmpl = """
        ProgramType         {algorithm}
        ObjectiveFunction   GCOP
        ModelExecutable     ./ostrich-runs-raven.sh
        PreserveBestModel   ./save_best.sh
        #OstrichWarmStart   yes

        ModelSubdir processor_

        BeginExtraDirs
        model
        #best
        EndExtraDirs

        BeginFilePairs
          {identifier}.rvi.tpl;  {identifier}.rvi
          {identifier}.rvp.tpl;  {identifier}.rvp
          {identifier}.rvc.tpl;  {identifier}.rvc
          {identifier}.rvt.tpl;  {identifier}.rvt
          #can be multiple (.rvh, .rvi)
        EndFilePairs

        #Parameter/DV Specification
        BeginParams
          #parameter init.    low               high               tx_in  tx_ost   tx_out
          par_x01    random   {lowerBounds.par_x01} {upperBounds.par_x01}  none   none     none
          par_x02    random   {lowerBounds.par_x02} {upperBounds.par_x02}  none   none     none
          par_x03    random   {lowerBounds.par_x03} {upperBounds.par_x03}  none   none     none
          par_x04    random   {lowerBounds.par_x04} {upperBounds.par_x04}  none   none     none
          par_x05    random   {lowerBounds.par_x05} {upperBounds.par_x05}  none   none     none
          par_x06    random   {lowerBounds.par_x06} {upperBounds.par_x06}  none   none     none
          par_x07    random   {lowerBounds.par_x07} {upperBounds.par_x07}  none   none     none
          par_x08    random   {lowerBounds.par_x08} {upperBounds.par_x08}  none   none     none
          par_x09    random   {lowerBounds.par_x09} {upperBounds.par_x09}  none   none     none
          par_x10    random   {lowerBounds.par_x10} {upperBounds.par_x10}  none   none     none
          par_x11    random   {lowerBounds.par_x11} {upperBounds.par_x11}  none   none     none
          par_x12    random   {lowerBounds.par_x12} {upperBounds.par_x12}  none   none     none
          par_x13    random   {lowerBounds.par_x13} {upperBounds.par_x13}  none   none     none
          par_x14    random   {lowerBounds.par_x14} {upperBounds.par_x14}  none   none     none
          par_x15    random   {lowerBounds.par_x15} {upperBounds.par_x15}  none   none     none
          par_x16    random   {lowerBounds.par_x16} {upperBounds.par_x16}  none   none     none
          par_x17    random   {lowerBounds.par_x17} {upperBounds.par_x17}  none   none     none
          par_x18    random   {lowerBounds.par_x18} {upperBounds.par_x18}  none   none     none
          par_x19    random   {lowerBounds.par_x19} {upperBounds.par_x19}  none   none     none
          par_x20    random   {lowerBounds.par_x20} {upperBounds.par_x20}  none   none     none
          par_x21    random   {lowerBounds.par_x21} {upperBounds.par_x21}  none   none     none
          par_x22    random   {lowerBounds.par_x22} {upperBounds.par_x22}  none   none     none
          par_x23    random   {lowerBounds.par_x23} {upperBounds.par_x23}  none   none     none
          par_x24    random   {lowerBounds.par_x24} {upperBounds.par_x24}  none   none     none
          par_x25    random   {lowerBounds.par_x25} {upperBounds.par_x25}  none   none     none
          par_x26    random   {lowerBounds.par_x26} {upperBounds.par_x26}  none   none     none
          par_x27    random   {lowerBounds.par_x27} {upperBounds.par_x27}  none   none     none
          par_x28    random   {lowerBounds.par_x28} {upperBounds.par_x28}  none   none     none
          par_x29    random   {lowerBounds.par_x29} {upperBounds.par_x29}  none   none     none
          par_x30    random   {lowerBounds.par_x30} {upperBounds.par_x30}  none   none     none
          par_x31    random   {lowerBounds.par_x31} {upperBounds.par_x31}  none   none     none
          par_x32    random   {lowerBounds.par_x32} {upperBounds.par_x32}  none   none     none
          par_x33    random   {lowerBounds.par_x33} {upperBounds.par_x33}  none   none     none
          par_x34    random   {lowerBounds.par_x34} {upperBounds.par_x34}  none   none     none
          par_x35    random   {lowerBounds.par_x35} {upperBounds.par_x35}  none   none     none
          par_r01    random   {lowerBounds.par_r01} {upperBounds.par_r01}  none   none     none
          par_r02    random   {lowerBounds.par_r02} {upperBounds.par_r02}  none   none     none
          par_r03    random   {lowerBounds.par_r03} {upperBounds.par_r03}  none   none     none
          par_r04    random   {lowerBounds.par_r04} {upperBounds.par_r04}  none   none     none
          par_r05    random   {lowerBounds.par_r05} {upperBounds.par_r05}  none   none     none
          par_r06    random   {lowerBounds.par_r06} {upperBounds.par_r06}  none   none     none
          par_r07    random   {lowerBounds.par_r07} {upperBounds.par_r07}  none   none     none
          par_r08    random   {lowerBounds.par_r08} {upperBounds.par_r08}  none   none     none
        EndParams

        BeginTiedParams
          # ---------------------------------------------------------------
          # MAX_MELT_FACTOR > MIN_MELT_FACTOR
          #
          # sum_x24_x25 = par_x24 + par_x25
          # Xtied =(c3 * X1 * X2) + (c2 * X2) + (c1 * X1) + c0
          # --> c0 = 0.0
          # --> c1 = 1.0
          # --> c2 = 1.0
          # --> c3 = 0.0
          #
          sum_x24_x25 2 par_x24 par_x25 linear 0.00 1.00 1.00 0.00 free
          #
          # ---------------------------------------------------------------
          # SNOW_SWI_MAX > SNOW_SWI_MIN
          #
          # sum_x13_x14 = par_x13 + par_x14
          # Xtied =(c3 * X1 * X2) + (c2 * X2) + (c1 * X1) + c0
          # --> c0 = 0.0
          # --> c1 = 1.0
          # --> c2 = 1.0
          # --> c3 = 0.0
          #
          sum_x13_x14 2 par_x13 par_x14 linear 0.00 1.00 1.00 0.00 free
          #
          # ---------------------------------------------------------------
          # FIELD_CAPACITY > SAT_WILT
          #
          # sum_x09_x10 = par_x09 + par_x10
          # Xtied =(c3 * X1 * X2) + (c2 * X2) + (c1 * X1) + c0
          # --> c0 = 0.0
          # --> c1 = 1.0
          # --> c2 = 1.0
          # --> c3 = 0.0
          #
          sum_x09_x10 2 par_x09 par_x10 linear 0.00 1.00 1.00 0.00 free
          #
          # ---------------------------------------------------------------
          # half the value but in [mm] not [m]
          #
          # half_x29 = par_x29 * 0.5 * 1000  --> half of it but in [mm] not [m]
          # Xtied = (c1 * X) + c0
          # --> c0 = 0.0
          # --> c1 = 500.
          #
          half_x29 1 par_x29 linear 500.0 0.0 free
          #
          # ---------------------------------------------------------------
          # half the value but in [mm] not [m]
          #
          # half_x30 = par_x30 * 0.5 * 1000  --> half of it but in [mm] not [m]
          # Xtied = (c1 * X) + c0
          # --> c0 = 0.0
          # --> c1 = 500.
          #
          half_x30 1 par_x30 linear 500.0 0.0 free
          #
          # ---------------------------------------------------------------
          # BASEFLOW_COEFF TOPSOIL  = 10.0^x4
          #
          # pow_x04 = 10.0**(par_x4)
          # Xtied = c2 * base ** (c1 * X) + c0
          # --> c0   = 0.0
          # --> c1   = 1.0
          # --> c2   = 1.0
          # --> base = 10.0
          #
          pow_x04 1 par_x04 exp 10.0 1.0 1.0 0.0 free
          #
          # ---------------------------------------------------------------
          # BASEFLOW_COEFF PHREATIC = 10.0^x11
          #
          # pow_x11 = 10.0**(par_x3)
          # Xtied = c2 * base ** (c1 * X) + c0
          # --> c0   = 0.0
          # --> c1   = 1.0
          # --> c2   = 1.0
          # --> base = 10.0
          #
          pow_x11 1 par_x11 exp 10.0 1.0 1.0 0.0 free
        EndTiedParams

        BeginResponseVars
          #name   filename                              keyword         line    col     token
          NS      ./model/output/{run_name}-{run_index}_Diagnostics.csv;       OST_NULL        1       3       ','
        EndResponseVars

        BeginTiedRespVars
          NegNS 1 NS wsum -1.00
        EndTiedRespVars

        BeginGCOP
          CostFunction NegNS
          PenaltyFunction APM
        EndGCOP

        BeginConstraints
                # not needed when no constraints, but PenaltyFunction statement above is required
                # name     type     penalty    lwr   upr   resp.var
        EndConstraints

        # Randomsed control added
        {random_seed}

        #Algorithm should be last in this file:

        BeginDDSAlg
                PerturbationValue 0.20
                MaxIterations {max_iterations}
                UseRandomParamValues
                # UseInitialParamValues
                # above intializes DDS to parameter values IN the initial model input files
        EndDDSAlg
        """
        self.config.ost.set_tmpl(ost_tmpl)

    def derived_parameters(self):
        """Derived parameters are computed by Ostrich."""
        self.config.rvt.rain_correction = "par_x33"
        self.config.rvt.snow_correction = "par_x34"

    def ost2raven(self, ops):
        """Return a list of parameter names calibrated by Ostrich that match Raven's parameters.

        Parameters
        ----------
        ops: dict
          Optimal parameter set returned by Ostrich.

        Returns
        -------
        BLENDEDParams named tuple
          Parameters expected by Raven.
        """
        names = ["par_x{:02}".format(i) for i in range(1, 36)] + [
            "par_r{:02}".format(i) for i in range(1, 9)
        ]
        names[3] = "pow_x04"
        names[9] = "sum_x09_x10"
        names[10] = "pow_x11"
        names[13] = "sum_x13_x14"
        names[24] = "sum_x24_x25"

        out = [ops[n] for n in names]
        return BLENDED.Params(*out)



# -*- coding: utf-8 -*-
"""
Created on Wed Jul 29 09:16:06 2015
@author: Michael Schramm
"""
from typing import Tuple

import numpy as np
from scipy.stats import norm

# TODO: This utility is written in python2 and will fail in python3 (e.g. no xrange)


def mk_test_calc(x: np.array, alpha: float = 0.05) -> Tuple[str, float, float, float]:
    """make test calculation.

    This function is derived from code originally posted by Sat Kumar Tomer
    (satkumartomer@gmail.com)
    See also: http://vsp.pnnl.gov/help/Vsample/Design_Trend_Mann_Kendall.htm
    The purpose of the Mann-Kendall (MK) test (Mann 1945, Kendall 1975, Gilbert
    1987) is to statistically assess if there is a monotonic upward or downward
    trend of the variable of interest over time. A monotonic upward (downward)
    trend means that the variable consistently increases (decreases) through
    time, but the trend may or may not be linear. The MK test can be used in
    place of a parametric linear regression analysis, which can be used to test
    if the slope of the estimated linear regression line is different from
    zero. The regression analysis requires that the residuals from the fitted
    regression line be normally distributed; an assumption not required by the
    MK test, that is, the MK test is a non-parametric (distribution-free) test.
    Hirsch, Slack and Smith (1982, page 107) indicate that the MK test is best
    viewed as an exploratory analysis and is most appropriately used to
    identify stations where changes are significant or of large magnitude and
    to quantify these findings.

    Parameters
    ----------
    x: np.array
      a vector of data.
    alpha: float
      significance level (0.05 default)

    Returns
    -------
    Tuple[str, float, float, float]

    Notes
    -----
    trend: tells the trend (increasing, decreasing or no trend)
    h: True (if trend is present) or False (if trend is absence)
    p: p value of the significance test
    z: normalized test statistics

    Examples
    --------
    >>> x = np.random.rand(100)
    >>> trend, h, p, z = mk_test(x,0.05)
    """
    n = len(x)

    # calculate S
    s = 0
    for k in range(n - 1):
        for j in range(k + 1, n):
            s += np.sign(x[j] - x[k])

    # calculate the unique data
    unique_x = np.unique(x)
    g = len(unique_x)

    # calculate the var(s)
    if n == g:  # there is no tie
        var_s = (n * (n - 1) * (2 * n + 5)) / 18
    else:  # there are some ties in data
        tp = np.zeros(unique_x.shape)
        for i in range(len(unique_x)):
            tp[i] = sum(x == unique_x[i])
        var_s = (n * (n - 1) * (2 * n + 5) - np.sum(tp * (tp - 1) * (2 * tp + 5))) / 18

    if s > 0:
        z = (s - 1) / np.sqrt(var_s)
    elif s < 0:
        z = (s + 1) / np.sqrt(var_s)
    else:  # s == 0:
        z = 0

    # calculate the p_value
    p = 2 * (1 - norm.cdf(abs(z)))  # two tail test
    h = abs(z) > norm.ppf(1 - alpha / 2)

    if (z < 0) and h:
        trend = "decreasing"
    elif (z > 0) and h:
        trend = "increasing"
    else:
        trend = "no trend"

    return trend, h, p, z


def check_num_samples(
    beta: float,
    delta: float,
    std_dev: float,
    alpha: float = 0.05,
    n: float = 4,
    num_iter: int = 1000,
    tol: float = 1e-6,
    num_cycles: int = 10000,
    m: int = 5,
) -> int:
    """Check number of samples.

    This function is an implementation of the "Calculation of Number of Samples
    Required to Detect a Trend" section written by Sat Kumar Tomer
    (satkumartomer@gmail.com) which can be found at:
    http://vsp.pnnl.gov/help/Vsample/Design_Trend_Mann_Kendall.htm
    As stated on the webpage in the URL above the method uses a Monte-Carlo
    simulation to determine the required number of points in time, n, to take a
    measurement in order to detect a linear trend for specified small
    probabilities that the MK test will make decision errors. If a non-linear
    trend is actually present, then the value of n computed by VSP is only an
    approximation to the correct n. If non-detects are expected in the
    resulting data, then the value of n computed by VSP is only an
    approximation to the correct n, and this approximation will tend to be less
    accurate as the number of non-detects increases.

    Parameters
    ----------
    beta: float
      probability of falsely accepting the null hypothesis
    delta: float
      change per sample period, i.e., the change that occurs between two adjacent sampling times
    std_dev: float
      standard deviation of the sample points.
    alpha: float
      significance level (0.05 default)
    n: int
      initial number of sample points (4 default).
    num_iter: int
      number of iterations of the Monte-Carlo simulation (1000 default).
    tol: float
      tolerance level to decide if the predicted probability is close enough to the required
      statistical power value (1e-6 default).
    num_cycles: int
      Total number of cycles of the simulation. This is to ensure that the simulation does finish
      regardless of convergence or not (10000 default).
    m: int
      if the tolerance is too small then the simulation could continue to cycle through the same sample
      numbers over and over. This parameter determines how many cycles to look back. If the same number of
      samples was been determined m cycles ago then the simulation will stop.

    Examples
    --------
    >>> num_samples = check_num_samples(0.2, 1, 0.1)
    """
    # Initialize the parameters
    power = 1.0 - beta
    p_d = 0.0
    cycle_num = 0
    min_diff_P_d_and_power = abs(p_d - power)
    best_P_d = p_d
    max_n = n
    min_n = n
    max_n_cycle = 1
    min_n_cycle = 1
    # Print information for user
    print(f"Delta (gradient): {delta}")
    print(f"Standard deviation: {std_dev}")
    print(f"Statistical power: {power}")

    # Compute an estimate of probability of detecting a trend if the estimate
    # Is not close enough to the specified statistical power value or if the
    # number of iterations exceeds the number of defined cycles.
    while abs(p_d - power) > tol and cycle_num < num_cycles:
        cycle_num += 1
        print(f"Cycle Number: {cycle_num}")
        count_of_trend_detections = 0

        # Perform MK test for random sample.
        for i in range(num_iter):
            r = np.random.normal(loc=0.0, scale=std_dev, size=n)
            x = r + delta * np.arange(n)
            trend, h, p, z = mk_test_calc(x, alpha)
            if h:
                count_of_trend_detections += 1
        p_d = float(count_of_trend_detections) / num_iter

        # Determine if p_d is close to the power value.
        if abs(p_d - power) < tol:
            print(f"P_d: {p_d}")
            print(f"{n} samples are required")
            return n

        # Determine if the calculated probability is closest to the statistical
        # power.
        if min_diff_P_d_and_power > abs(p_d - power):
            min_diff_P_d_and_power = abs(p_d - power)
            best_P_d = p_d

        # Update max or min n.
        if n > max_n and abs(best_P_d - p_d) < tol:
            max_n = n
            max_n_cycle = cycle_num
        elif n < min_n and abs(best_P_d - p_d) < tol:
            min_n = n
            min_n_cycle = cycle_num

        # In case the tolerance is too small we'll stop the cycling when the
        # number of cycles, n, is cycling between the same values.
        elif (
            abs(max_n - n) == 0
            and cycle_num - max_n_cycle >= m
            or abs(min_n - n) == 0
            and cycle_num - min_n_cycle >= m
        ):
            print("Number of samples required has converged.")
            print(f"P_d: {p_d}")
            print(f"Approximately {n} samples are required")
            return n

        # Determine whether to increase or decrease the number of samples.
        if p_d < power:
            n += 1
            print(f"P_d: {p_d}")
            print(f"Increasing n to {n}")
            print("")
        else:
            n -= 1
            print(f"P_d: {p_d}")
            print(f"Decreasing n to {n}")
            print("")
            if n == 0:
                raise ValueError("Number of samples = 0. This should not happen.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 23 22:25:32 2021

@author: ets
"""

import datetime as dt
import logging

from ravenpy.models.emulators import GR4JCN
from ravenpy.utilities.forecasting import (
    make_climpred_hindcast_object,
    make_ESP_hindcast_dataset,
)
from ravenpy.utilities.testdata import get_local_testdata


class TestClimpredHindcastVerification:
    def test_simple(self):

        # We don't want climpred logging in the test output
        logger = logging.getLogger()
        logger.setLevel(logging.WARNING)

        # Prepare the model parameters and forecast details
        model = "GR4JCN"
        params = (0.529, -3.396, 407.29, 1.072, 16.9, 0.947)

        forecast_duration = 3
        ts = get_local_testdata(
            "raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc"
        )
        rvc = get_local_testdata("gr4j_cemaneige/solution.rvc")

        # Make the hindcasts for each initialization date. Here we will extract
        # ESP forecasts for a given calendar date for the years in "included_years"
        # as hindcast dates. Each ESP hindcast uses all available data in the ts dataset,
        # so in this case we will have 56/57 members for each hindcast initialization
        # depending on the date that we start on. The "hindcasts" dataset contains
        # all of the flow data from the ESP hindcasts for the initialization dates.
        # The "qobs" dataset contains all qobs in the timeseries: Climpred will
        # sort it all out during its processing. Note that the format of these datasets
        # is tailor-made to be used in climpred, and thus has specific dimension names.

        hindcasts, qobs = make_ESP_hindcast_dataset(
            model_name=model,
            forecast_date=dt.datetime(1955, 6, 30),
            included_years=list(range(1957, 1959)),
            forecast_duration=forecast_duration,
            ts=ts,
            hrus=(
                GR4JCN.LandHRU(
                    area=4250.6, elevation=843.0, latitude=54.4848, longitude=-123.3659
                ),
            ),
            params=params,
            rvc=str(rvc),
        )

        # Once we have the correctly formatted datasets, Make the hindcast object for climpred
        hindcast_object = make_climpred_hindcast_object(hindcasts, qobs)

        # This function is used to convert to binary to see if yes/no forecast is larger than obs
        def pos(x):
            return x > 0  # Check for binary outcome

        # These three functions respectively compute the rank histogram,
        # the crps and the reliability for the set of initialized dates
        # (i.e. forecast issue dates, here 1 day per year at the same calendar day).
        rank_histo_verif = hindcast_object.verify(
            metric="rank_histogram",
            comparison="m2o",
            dim=["member", "init"],
            alignment="same_inits",
        )
        crps_verif = hindcast_object.verify(
            metric="crps",
            comparison="m2o",
            dim=["member", "init"],
            alignment="same_inits",
        )
        reliability_verif = hindcast_object.verify(
            metric="reliability",
            comparison="m2o",
            dim=["member", "init"],
            alignment="same_inits",
            logical=pos,
        )

        assert "flow" in rank_histo_verif
        assert "flow" in crps_verif
        assert "flow" in reliability_verif
        assert rank_histo_verif.flow.shape[0] == forecast_duration
        assert reliability_verif.flow.shape[0] == forecast_duration
        assert crps_verif.flow.shape[0] == forecast_duration



"""
Base classes
------------

The `Raven` class is the base class that implements model setup, execution and output retrieval, while the `Ostrich`
class is the base class adapting `Raven` to work with the Ostrich calibration tool.

"""
import csv
import datetime as dt
import operator
import os
import re
import shutil
import stat
import subprocess
import tempfile
from collections import OrderedDict
from dataclasses import astuple, fields, is_dataclass, replace
from pathlib import Path
from typing import Union

import numpy as np
import xarray as xr
from numpy.distutils.misc_util import is_sequence

import ravenpy
from ravenpy.config.commands import (
    DataCommand,
    GriddedForcingCommand,
    HRUsCommand,
    ObservationDataCommand,
    StationForcingCommand,
)
from ravenpy.config.rvs import RVC, RVI, Config

RAVEN_EXEC_PATH = os.getenv("RAVENPY_RAVEN_BINARY_PATH") or shutil.which("raven")
OSTRICH_EXEC_PATH = os.getenv("RAVENPY_OSTRICH_BINARY_PATH") or shutil.which("ostrich")

RAVEN_NO_DATA_VALUE = -1.2345


class RavenError(Exception):
    """
    This is an error that is meant to be raised whenever a message of type "ERROR" is found
    in the Raven_errors.txt file resulting from a Raven (i.e. the C program) run.
    """

    pass


class Raven:
    """RAVEN hydrological model wrapper.

    This class is used to run the RAVEN model from user-provided configuration files. It can also be subclassed with
    configuration templates for emulated models, allowing direct calls to the models.

    r = Raven('/tmp/testdir')
    r.configure()
    """

    _parallel_parameters = [
        "params",
        "hru_state",
        "basin_state",
        "nc_index",
        # "name",
        "area",
        "elevation",
        "latitude",
        "longitude",
        "region_id",
    ]

    def __init__(self, workdir: Union[str, Path] = None):  # , identifier: str = None):
        """Initialize the RAVEN model.

        Directory for the model configuration and outputs. If None, a temporary directory will be created.
        """

        if not RAVEN_EXEC_PATH:
            raise RuntimeError(
                "Could not find raven binary in PATH, and RAVENPY_RAVEN_BINARY_PATH env variable is not set"
            )

        if not OSTRICH_EXEC_PATH:
            raise RuntimeError(
                "Could not find ostrich binary in PATH, and RAVENPY_OSTRICH_BINARY_PATH env variable is not set"
            )

        self.raven_exec = RAVEN_EXEC_PATH
        self.ostrich_exec = OSTRICH_EXEC_PATH

        workdir = workdir or tempfile.mkdtemp()

        self.workdir = Path(workdir)
        self.ind_outputs = {}  # Individual files for all simulations
        self.outputs = {}  # Aggregated files
        self.singularity = False  # Set to True to launch Raven with singularity.
        self.raven_simg = None  # ravenpy.raven_simg
        # self._name = None
        self._defaults = {}

        self._rv_paths = []

        # Directory logic
        # Top directory inside workdir. This is where Ostrich and its config and templates are stored.
        self.model_dir = "model"  # Path to the model configuration files.
        self.final_dir = "final"
        self.output_dir = "output"

        self.exec_path = self.workdir / "exec"
        self.final_path = self.workdir / self.final_dir
        self._psim = 0
        self._pdim = None  # Parallel dimension (either initparam, params or region)

        self.config = Config(raven_version=self.version)

    @property
    def output_path(self):
        return self.model_path / self.output_dir

    @property
    def model_path(self):
        return self.exec_path / self.model_dir / "p{:02}".format(self.psim)

    @property
    def raven_cmd(self):
        """Path to the Raven executable."""
        return self.model_path / "raven"

    @property
    def version(self):
        import re

        out = subprocess.check_output(
            [
                self.raven_exec,
            ],
            input=b"\n",
        )
        match = re.search(r"Version (\S+) ", out.decode("utf-8"))
        if match:
            return match.groups()[0]
        else:
            raise AttributeError("Version not found: {}".format(out))

    @property
    def psim(self):
        return self._psim

    @psim.setter
    def psim(self, value):
        if not isinstance(value, int):
            raise ValueError
        if isinstance(self.config.rvi, RVI):
            self.config.rvi.run_index = value
        self._psim = value

    @property
    def cmd(self):
        """This is the main executable."""
        return self.raven_cmd

    @property
    def bash_cmd(self):
        """Bash command arguments."""
        identifier = self.config.identifier or "raven-generic"
        return [self.cmd, identifier, "-o", str(self.output_path)]

    @property
    def singularity_cmd(self):
        """Run Singularity container."""
        return [
            "singularity",
            "run",
            "--bind",
            "{}:/data".format(self.model_path),
            "--bind",
            "{}:/data_out:rw".format(self.output_path),
            self.raven_simg,
            self.config.identifier,
        ]

    @property
    def cmd_path(self):
        """This is the main executable."""
        return self.model_path

    def derived_parameters(self):
        """Subclassed by emulators. Defines model parameters that are a function of other parameters."""
        return

    def configure(self, fns):
        """Read configuration files."""
        if not is_sequence(fns):
            fns = [fns]
        for fn in map(Path, fns):
            self.config.set_rv_file(fn)

    def _dump_rv(self):
        """Write configuration files to disk."""

        identifier = self.config.identifier or "raven-generic"

        for rvx in ["rvt", "rvh", "rvp", "rvc", "rvi"]:
            rvo = getattr(self.config, rvx)
            if rvo.is_ostrich_tmpl:
                fn = self.exec_path / f"{identifier}.{rvx}.tpl"
            else:
                fn = self.model_path / f"{identifier}.{rvx}"
            with open(fn, "w") as f:
                self._rv_paths.append(fn)
                content = rvo.content or rvo.to_rv()
                assert content.strip(), f"{rvx} has no content!"
                f.write(content)

    def setup(self, overwrite=False):
        """Create directory structure to store model input files, executable and output results.

        Model configuration files and time series inputs are stored directly in the working directory.

        workdir/  # Created by PyWPS. Is considered the model path.
        model/
        output/

        """
        if overwrite:
            if self.model_path.exists():
                shutil.rmtree(str(self.exec_path))
            if self.final_path.exists():
                shutil.rmtree(str(self.final_path))

        # Create general subdirectories
        if not self.exec_path.exists():
            os.makedirs(str(self.exec_path))  # workdir/exec
        if not self.final_path.exists():
            os.makedirs(str(self.final_path))  # workdir/final

    def setup_model_run(self, ts):
        """Create directory structure to store model input files, executable and output results.

        Parameters
        ----------
        ts : sequence
          Paths to input forcing files.
        index : int
          Run index (starts at 1)
        """
        # Create configuration information from input files
        # ncvars = self._assign_files(ts)
        # self.rvt.update(ncvars)

        # TODO: Re-enable those checks
        # self.check_units()
        # self.check_inputs()

        # Compute derived parameters
        self.derived_parameters()

        # Write configuration files in model directory
        if not self.model_path.exists():
            os.makedirs(self.model_path)
            os.makedirs(self.output_path)
        self._dump_rv()

        # Create symbolic link to input files
        for fn in ts:
            if not (self.model_path / Path(fn).name).exists():
                os.symlink(str(fn), str(self.model_path / Path(fn).name))

        # Create symbolic link to Raven executable
        if not self.raven_cmd.exists():
            os.symlink(self.raven_exec, str(self.raven_cmd))

        # Shell command to run the model
        if self.singularity:
            cmd = self.singularity_cmd
        else:
            cmd = self.bash_cmd

        return cmd

    def run(self, ts, overwrite=False, **kwds):
        """Run the model.

        Parameters
        ----------
        ts : path or sequence
          Sequence of input file paths. Symbolic links to those files will be created in the model directory.
        overwrite : bool
          Whether or not to overwrite existing model and output files.
        **kwds : dict
          Raven parameters used to fill configuration file templates.

        Create a work directory with a model/ and output/ subdirectories, write the configuration files in model/ and
        launch the Raven executable. If the configuration files are templates, values can be formatted by passing
        dictionaries keyed by their extension.

        Examples
        --------
        >>> r = Raven()
        >>> r.configure(rvi='path to template', rvp='...'}
        >>> r.run(ts, start_date=dt.datetime(2000, 1, 1), area=1000, X1=67)

        """
        if isinstance(ts, (str, Path)):
            ts = [ts]

        ts_are_ncs = all(Path(f).suffix.startswith(".nc") for f in ts)

        # Support legacy interface for single HRU emulator
        hru_attrs = {}
        for k in ["area", "latitude", "longitude", "elevation"]:
            v = kwds.pop(k, None)
            if v:
                # It seems that `v` is a list when running via a WPS interface
                hru_attrs[k] = v[0] if isinstance(v, list) else v
        if hru_attrs:
            assert len(self.config.rvh.hrus) == 1
            self.config.rvh.hrus = (replace(self.config.rvh.hrus[0], **hru_attrs),)

        # Case for potentially parallel parameters
        pdict = {}
        for p in self._parallel_parameters:
            a = kwds.pop(p, None)

            if (
                a is not None
                and p == "params"
                and not isinstance(a, self.__class__.Params)
            ):
                pdict[p] = np.atleast_2d(a)
            else:
                pdict[p] = np.atleast_1d(a)

        # Number of parallel loops is dictated by the number of parallel parameters or nc_index.
        plen = {pp: len(pdict[pp]) for pp in self._parallel_parameters + ["nc_index"]}

        # Find the longest parallel array and its length
        longer, nloops = max(plen.items(), key=operator.itemgetter(1))

        # Assign the name of the parallel dimension
        # nbasins is set by RavenC++
        if nloops > 1:
            self._pdim = {
                "params": "params",
                "hru_state": "state",
                "basin_state": "state",
                "nc_index": "nbasins",
            }[longer]

        for key, val in pdict.items():
            if len(val) not in [1, nloops]:
                raise ValueError(
                    "Parameter {} has incompatible dimension: {}. "
                    "Should be 1 or {}.".format(key, len(val), nloops)
                )

        # Resize parallel parameters to the largest size
        for key, val in pdict.items():
            if len(val) == 1:
                pdict[key] = val.repeat(nloops, axis=0)

        # Use rvc file to set model state, if any
        rvc = kwds.pop("rvc", None)
        if rvc:
            self.resume(solution=rvc)

        # Update non-parallel parameter objects
        for key, val in kwds.items():
            self.config.update(key, val)

        # if self.config.rvi:
        if ts_are_ncs:
            self.handle_date_defaults(ts)
            self.set_calendar(ts)

        if ts_are_ncs:
            self.config.rvt.configure_from_nc_data(ts)

        # Loop over parallel parameters - sets self.rvi.run_index
        procs = []
        for self.psim in range(nloops):
            for key, val in pdict.items():
                if val[self.psim] is not None:
                    if key == "hru_state":
                        self.config.rvc.set_hru_state(val[self.psim])
                    elif key == "basin_state":
                        self.config.rvc.set_basin_state(val[self.psim])
                    else:
                        self.config.update(key, val[self.psim])

            cmd = self.setup_model_run(tuple(map(Path, ts)))

            procs.append(
                subprocess.Popen(
                    cmd,
                    cwd=self.cmd_path,
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    universal_newlines=True,
                )
            )

        return procs

    def __call__(self, ts, overwrite=False, **kwds):
        self.setup(overwrite)

        procs = self.run(ts, overwrite, **kwds)

        for proc in procs:
            # When Raven errors right away (for instance if it's missing an RV file)
            # it asks for a RETURN to exit
            proc.communicate(input="\n")
            proc.wait()

        messages = self.extract_raven_messages()

        if messages["ERROR"]:
            raise RavenError("\n".join(messages["ERROR"]))

        assert messages["SIMULATION COMPLETE"]

        self.parse_results()

    def resume(self, solution=None):
        """Set the initial state to the state at the end of the last run.

        Parameters
        ----------
        solution : str, Path
          Path to solution file. If None, will use solution from last model run if any.
        """
        if solution is None:
            fn = self.outputs["solution"]
        else:
            fn = solution

        self.config.rvc.parse_solution(Path(fn).read_text())

    def parse_results(self, path=None, run_name=None):
        """Store output files in the self.outputs dictionary."""
        # Output files default names. The actual output file names will be composed of the run_name and the default
        # name.
        path = path or self.exec_path
        run_name = run_name or self.config.rvi.run_name or ""
        patterns = {
            "hydrograph": f"{run_name}*Hydrographs.nc",
            "storage": f"{run_name}*WatershedStorage.nc",
            "solution": f"{run_name}*solution.rvc",
            "diagnostics": f"{run_name}*Diagnostics.csv",
        }

        for key, pattern in patterns.items():
            # There are no diagnostics if a streamflow time series is not provided.
            try:
                fns = self._get_output(pattern, path=path)
            except UserWarning as exc:
                if key != "diagnostics":
                    raise exc
                else:
                    continue

            fns.sort()
            self.ind_outputs[key] = fns
            self.outputs[key] = self._merge_output(fns, pattern[1:])

        self.outputs["rv_config"] = self._merge_output(self._rv_paths, "rv.zip")

    def _merge_output(self, files, name):
        """Merge multiple output files into one if possible, otherwise return a list of files."""
        import zipfile

        # If there is only one file, return its name directly.
        if len(files) == 1:
            return files[0]

        # Otherwise try to create a new file aggregating all files.
        outfn = self.final_path / name

        if name.endswith(".nc") and not isinstance(
            self, ravenpy.models.RavenMultiModel
        ):
            ds = [xr.open_dataset(fn) for fn in files]
            try:
                # We aggregate along the pdim dimensions.
                out = xr.concat(ds, self._pdim, data_vars="all")
                out.to_netcdf(outfn)
                return outfn
            except (ValueError, KeyError):
                pass

        # Let's zip the files that could not be merged.
        outfn = outfn.with_suffix(".zip")

        # Find the lower file parts level at which there are differences among files.
        i = get_diff_level(files)

        # Try to create a zip file
        with zipfile.ZipFile(outfn, "w") as f:
            for fn in files:
                len(fn.parts)
                f.write(fn, arcname=fn.relative_to(Path(*fn.parts[:i])))

        return outfn

    def extract_raven_messages(self):
        """
        Parse all the Raven_errors and extract the messages, structured by types.
        """
        err_filepaths = self.exec_path.rglob("Raven_errors.txt")
        messages = {
            "ERROR": [],
            "WARNING": [],
            "ADVISORY": [],
            "SIMULATION COMPLETE": False,
        }
        for p in err_filepaths:
            # The error message for an unknown command is exceptionally on two lines
            # (the second starts with a triple space)
            for m in re.findall("^([A-Z ]+) :(.+)(?:\n   (.+))?", p.read_text(), re.M):
                if m[0] == "SIMULATION COMPLETE":
                    messages["SIMULATION COMPLETE"] = True
                    continue
                msg_type = m[0]
                msg = f"{m[1]} {m[2]}".strip()
                if (
                    msg
                    == "Errors found in input data. See Raven_errors.txt for details"
                ):
                    # Skip this one because it's a bit circular
                    continue
                messages[msg_type].append(msg)

        return messages

    def _get_output(self, pattern, path):
        """Match actual output files to known expected files.

        Return a dictionary of file paths for each expected input.
        """
        files = list(path.rglob(pattern))

        if len(files) == 0:
            if not (
                isinstance(self.config.rvi, RVI) and self.config.rvi.suppress_output
            ):
                raise UserWarning("No output files for {} in {}.".format(pattern, path))

        return [f.absolute() for f in files]

    @staticmethod
    def start_end_date(fns):
        """Return the common starting and ending date and time of netCDF files.

        Parameters
        ----------
        fns : sequence
          Sequence of netCDF file names for forcing data.

        Returns
        -------
        start : datetime
          The first datetime of the forcing files.
        end : datetime
          The last datetime of the forcing files.
        """

        ds = xr.open_mfdataset(fns, combine="by_coords")
        return ds.indexes["time"][0], ds.indexes["time"][-1]

    @staticmethod
    def get_calendar(fns):
        """Return the calendar."""
        ds = xr.open_mfdataset(fns, combine="by_coords")
        cal = ds.time.encoding.get("calendar", "standard")
        return RVI.CalendarOptions(cal.upper())

    def set_calendar(self, ts):
        """Set the calendar in the RVI configuration."""
        self.config.rvi.calendar = self.get_calendar(ts)

    def handle_date_defaults(self, ts):
        # Get start and end date from file
        start, end = self.start_end_date(ts)

        rvi = self.config.rvi
        if rvi.start_date in [None, dt.datetime(1, 1, 1)]:
            rvi.start_date = start

        if rvi.end_date in [None, dt.datetime(1, 1, 1)]:
            rvi.end_date = end

    @property
    def q_sim(self):
        """Return a view of the hydrograph time series.

        This view will be overwritten by successive calls to `run`. To make a copy of this DataArray that will
        persist in memory, use `q_sim.copy(deep=True)`.
        """
        if isinstance(self.hydrograph, list):
            return [h.q_sim for h in self.hydrograph]

        return self.hydrograph.q_sim

    @property
    def hydrograph(self):
        """Return a view of the current output file.

        If the model is run multiple times, hydrograph will point to the latest version. To store the results of
        multiple runs, either create different model instances or explicitly copy the file to another disk location.
        """
        if self.outputs["hydrograph"].suffix == ".nc":
            return xr.open_dataset(self.outputs["hydrograph"])
        elif self.outputs["hydrograph"].suffix == ".zip":
            return [xr.open_dataset(fn) for fn in self.ind_outputs["hydrograph"]]
        else:
            raise ValueError

    @property
    def storage(self):
        if self.outputs["storage"].suffix == ".nc":
            return xr.open_dataset(self.outputs["storage"])
        elif self.outputs["storage"].suffix == ".zip":
            return [xr.open_dataset(fn) for fn in self.ind_outputs["storage"]]
        else:
            raise ValueError

    @property
    def solution(self):
        if self.outputs["solution"].suffix == ".rvc":
            return RVC.create_solution(self.outputs["solution"].read_text())
        elif self.outputs["solution"].suffix == ".zip":
            return [
                RVC.create_solution(fn.read_text())
                for fn in self.ind_outputs["solution"]
            ]

    def get_final_state(self, hru_index=1, basin_index=1):
        """Return model state at the end of simulation.

        Parameters
        ----------
        hru_index : None, int
          Set index value or None to get all HRUs.
        basin_index : None, int
          Set index value or None to get all basin states.
        """
        solution = self.solution
        if isinstance(solution, RVC):
            return solution.hru_states[hru_index], solution.basin_states[basin_index]
        else:
            states = [
                (sol.hru_states[hru_index], sol.basin_states[basin_index])
                for sol in solution
            ]
            return zip(*states)

    @property
    def diagnostics(self):
        diag = []
        for fn in self.ind_outputs["diagnostics"]:
            with open(fn) as f:
                reader = csv.reader(f.readlines())
                header = next(reader)
                content = next(reader)

                out = dict(zip(header, content))
                out.pop("")

            for key, val in out.items():
                if "DIAG" in key:
                    out[key] = float(val)
            diag.append(out)

        return diag if len(diag) > 1 else diag[0]


class Ostrich(Raven):
    """Wrapper for OSTRICH calibration of RAVEN hydrological model.

    This class is used to calibrate RAVEN model using OSTRICH from user-provided configuration files. It can also be
    subclassed with configuration templates for emulated models, allowing direct calls to the models.

    Parameters
    ----------
    conf:
      The rv configuration files + Ostrict ostIn.txt.
    tpl:
      The Ostrich templates.

    Examples
    --------
    >>> r = Ostrich('/tmp/testdir')
    >>> r.configure()
    """

    @property
    def model_path(self):
        return self.exec_path / self.model_dir

    @staticmethod
    def _allowed_extensions():
        return Raven._allowed_extensions() + ("txt",)

    @property
    def ostrich_cmd(self):
        """OSTRICH executable path."""
        return self.exec_path / "ostrich"

    @property
    def cmd(self):
        """OSTRICH executable path."""
        return self.ostrich_cmd

    @property
    def cmd_path(self):
        """This is the main executable."""
        return self.exec_path

    @property
    def proc_path(self):
        """Path to Ostrich parallel process directory."""
        return self.exec_path / "processor_0"  # /'model' / 'output' ?

    def write_save_best(self):
        fn = self.exec_path / "save_best.sh"
        fn.write_text(save_best)
        make_executable(fn)

    def write_ostrich_runs_raven(self):
        fn = self.exec_path / "ostrich-runs-raven.sh"
        identifier = self.config.identifier or "ostrich-generic"
        fn.write_text(ostrich_runs_raven.format(identifier=identifier))
        make_executable(fn)

    def setup(self, overwrite=False):
        """Create directory structure to store model input files, executable, and output results.

        Model configuration files and time series inputs are stored directly in the working directory.
        At each Ostrich loop, configuration files (original and created from templates are copied into model).
        """
        Raven.setup(self, overwrite)

        os.makedirs(str(self.final_path), exist_ok=True)

        self.write_ostrich_runs_raven()
        self.write_save_best()

        # Create symbolic link to executable
        os.symlink(self.ostrich_exec, str(self.cmd))

    def _dump_rv(self):
        """write configuration files to disk."""

        super()._dump_rv()

        # ostIn.txt
        fn = self.exec_path / "ostIn.txt"
        with open(fn, "w") as f:
            self._rv_paths.append(fn)
            f.write(self.config.ost.content or self.config.ost.to_rv())

        # OstRandomNumbers.txt
        if self.config.ost.random_numbers_path:
            fn = self.exec_path / "OstRandomNumbers.txt"
            with open(fn, "w") as f:
                self._rv_paths.append(fn)
                f.write(self.config.ost.random_numbers_path.read_text())

    def parse_results(self):
        """Store output files in the self.outputs dictionary."""
        # Output files default names. The actual output file names will be composed of the run_name and the default
        # name.
        Raven.parse_results(self, path=self.final_path)

        patterns = {
            "params_seq": "OstModel?.txt",
            "calibration": "OstOutput?.txt",
        }

        # Store output file names in dict
        for key, pattern in patterns.items():
            fns = self._get_output(pattern, path=self.exec_path)
            if len(fns) == 1:
                fns = fns[0]
            self.outputs[key] = fns

        try:
            if is_dataclass(self.calibrated_params):
                self.outputs["calibparams"] = ", ".join(
                    map(str, astuple(self.calibrated_params))
                )
            else:
                self.outputs["calibparams"] = ", ".join(
                    map(str, self.calibrated_params)
                )
        except (AttributeError, TypeError):
            err = self.parse_errors()
            raise UserWarning(err)

    def parse_errors(self):
        try:
            raven_err = self._get_output("OstExeOut.txt", path=self.exec_path)[
                0
            ].read_text()
        except UserWarning:  # Read in processor_0 directory instead.
            try:
                raven_err = self._get_output("OstExeOut.txt", path=self.proc_path)[
                    0
                ].read_text()
            except UserWarning:
                raven_err = ""

        try:
            ost_err = self._get_output("OstErrors?.txt", path=self.exec_path)[
                0
            ].read_text()
        except UserWarning:  # Read in processor_0 directory instead.
            ost_err = self._get_output("OstErrors?.txt", path=self.proc_path)[
                0
            ].read_text()

        return f"{ost_err}\n{raven_err}"

    def parse_optimal_parameter_set(self):
        """Return dictionary of optimal parameter set."""
        import re

        txt = open(self.outputs["calibration"]).read()
        ops = re.search(r".*Optimal Parameter Set(.*?)\n{2}", txt, re.DOTALL).groups()[
            0
        ]

        p = re.findall(r"(\w+)\s*:\s*([\S]+)", ops)
        return OrderedDict((k, float(v)) for k, v in p)

    def ost2raven(self, ops):
        """Return model parameters.

        Notes
        -----
        This method should be subclassed by emulators for which Ostrich has different parameters than the original
        Raven model.
        """

        if self.config.rvp.params:
            # We are using an emulator
            n = len(fields(self.config.rvp.params))
            pattern = "par_x{}" if n < 8 else "par_x{:02}"
            names = [pattern.format(i + 1) for i in range(n)]
            return self.__class__.Params(*[ops[n] for n in names])
        else:
            # We are using generic Ostrich
            return ops.values()

    @property
    def calibrated_params(self):
        """The dictionary of optimal parameters estimated by Ostrich."""
        ops = self.parse_optimal_parameter_set()
        return self.ost2raven(ops)

    @property
    def obj_func(self):
        return np.loadtxt(self.outputs["params_seq"], skiprows=1)[-1, 1]

    @property
    def optimized_parameters(self):
        """These are the raw parameters returned by Ostrich."""
        return np.loadtxt(self.outputs["params_seq"], skiprows=1)[-1, 2:]


def get_diff_level(files):
    """Return the lowest hierarchical file parts level at which there are differences among file paths."""

    for i, parts in enumerate(zip(*[f.parts for f in files])):
        if len(set(parts)) > 1:
            return i


def make_executable(fn):
    """Make file executable."""
    st = os.stat(fn)
    os.chmod(fn, st.st_mode | stat.S_IEXEC)


def get_average_annual_runoff(
    nc_file_path, area_in_m2, time_dim="time", na_value=RAVEN_NO_DATA_VALUE
):
    """
    Compute the average annual runoff from observed data.
    """
    with xr.open_dataset(nc_file_path) as ds:
        qobs = ds.where(ds["qobs"] != na_value)["qobs"]
        qobs *= 86400.0  # convert m**3/s to m**3/d
        axis = qobs.dims.index(time_dim)
        # avg daily runoff [m3/d] for each year in record
        qyear = np.nanmean(qobs.groupby("time.year").mean("time"), axis=axis)
        qyear = qyear / area_in_m2 * 365 * 1000.0  # [mm/yr] for each year in record
        qyear = np.mean(qyear)  # [mm/yr] mean over all years in record

    return qyear


# TODO: Configure this according to the model_path and output_path.
save_best = """#!/bin/bash

set -e

cp ./model/*.rv?  ../../final/
cp ./model/output/* ../../final/

exit 0
"""

# TODO: Configure this according to raven_cmd, name and output_path.
ostrich_runs_raven = """
#!/bin/bash

set -e

cp ./*.rv? model/

./model/raven ./model/{identifier} -o ./model/output/

exit 0
"""



"""
This module contains functions creating web-friendly interactive graphics using holoviews.

The graphic outputs are meant to be displayed in a notebook.
In a console, use `hvplot.show(fig)` to render the figures.
"""
from pathlib import Path

import numpy as np
import xarray as xr
from matplotlib import pyplot as plt

try:
    import holoviews as hv
    import hvplot.xarray

    # from holoviews.streams import Buffer
    # from bokeh.models import Range1d, LinearAxis
except (ImportError, ModuleNotFoundError) as e:
    raise ImportError("Install holoviews and hvplot.") from e

hv.extension("bokeh")


def hydrographs(ds):
    """Return a graphic showing the discharge simulations and observations."""

    basin_name = ds.basin_name.values[0]  # selected basin name
    g = ds.q_sim.hvplot.line(
        x="time",
        line_width=1,
        label="Simulations",
        ylabel="Streamflow (m³/s)",
        xlabel="Time",
        title=basin_name,
    )

    # Plot the observed streamflows if available
    if hasattr(ds, "q_obs"):
        g *= ds.q_obs.hvplot.line(
            x="time", line_width=1.5, color="k", label="Observations"
        )

    return g


def mean_annual_hydrograph(ds):
    """Return a graphic showing the discharge simulations and observations."""

    basin_name = ds.basin_name.values[0]  # selected basin name
    mq_sim = ds.q_sim.groupby("time.dayofyear").mean()

    g = mq_sim.hvplot.line(
        x="dayofyear",
        line_width=1.5,
        label="Mean simulation",
        ylabel="Mean streamflow (m³/s)",
        xlabel="Day of year",
        title=basin_name,
    )

    # Plot the observed streamflows if available
    if hasattr(ds, "q_obs"):
        mq_obs = ds.q_obs.groupby("time.dayofyear").mean()
        g *= mq_obs.hvplot.line(
            x="dayofyear", line_width=2, color="k", label="Mean observations"
        )

    return g


def spaghetti_annual_hydrograph(ds):
    """
    Create a spaghetti plot of the mean hydrological cycle for one model
    simulations.
    """
    basin_name = ds.basin_name.values[0]  # selected basin name
    mq_sim = ds.q_sim.groupby("time.dayofyear").mean()
    g1 = mq_sim.hvplot.line(
        x="dayofyear",
        line_width=1.5,
        label="Mean simulation",
        ylabel="Mean streamflow (m³/s)",
        xlabel="Day of year",
        title=basin_name,
    )

    g1 *= ds.q_sim.hvplot.line(
        x="time.dayofyear",
        by="time.year",
        line_width=1,
        label="Simulations",
        legend=False,
    )

    # Plot the observed streamflows if available
    if hasattr(ds, "q_obs"):
        mq_obs = ds.q_obs.groupby("time.dayofyear").mean()
        g2 = mq_obs.hvplot.line(
            x="dayofyear", line_width=2, color="k", label="Mean observations"
        )

        g2 *= ds.q_obs.hvplot.line(
            x="time.dayofyear",
            by="time.year",
            line_width=1,
            color="k",
            label="Observations",
            legend=False,
        )

    return g1 + g2


def ts_fit_graph(ts, params):
    """Create graphic showing an histogram of the data and the distribution fitted to it.

    The graphic contains one panel per watershed.

    Parameters
    ----------
    ts : xr.DataArray
      Stream flow time series with dimensions (time, nbasins).
    params : xr.DataArray
      Fitted distribution parameters returned by `xclim.land.fit` indicator.

    Returns
    -------
    fig
      Figure showing a histogram and the parameterized pdf.
    """
    # Note: The hover tool could be customized to show the histogram count in addition to the frequency.
    from xclim.indices.stats import get_dist

    n = ts.nbasins.size
    if n > 1:
        raise NotImplementedError

    ts = ts.isel(nbasins=0)
    params = params.isel(nbasins=0)

    # Using matplotlib's default binning strategy
    hist, bins, mh = plt.hist(ts, bins="auto", density=True)

    # Histogram graphic object
    h = hv.Histogram((hist, bins), kdims=ts.name, label="Histogram")

    # PDF domain
    mn = np.min(bins)
    mx = np.max(bins)
    q = np.linspace(mn, mx, 200)

    # Compute PDF
    dist = params.attrs["scipy_dist"]
    dc = get_dist(dist)(*params)  # Works because dparams is the first dimension.
    pdf = xr.DataArray(
        data=dc.pdf(q),
        dims=(ts.name,) + params.dims[1:],
        coords={ts.name: q},
        name="pdf",
    )

    # PDF line label
    ps = ", ".join(
        [
            f"{key}={x:.1f}".format(x)
            for (key, x) in zip(params.dparams.data, params.values)
        ]
    )
    label = "{}({})".format(dist, ps)

    # PDF graphic object
    p = pdf.hvplot.line(label=label, xlabel=ts.attrs["long_name"], color="orange")

    # Layout
    return (h * p).opts(hv.opts.Histogram(tools=["hover"]))
